!=========================================================================
! Copyright (C) GemStone Systems, Inc. 1986-2011.  All Rights Reserved.
!
! $Id: btreeleaf.gs,v 1.7 2008-01-09 19:02:30 stever Exp $
!
! Superclass Hierarchy:
!   BtreeLeafNode, BtreeNode, Array, SequenceableCollection, Collection,
!   Object.
!
!=========================================================================

! class created in idxclasses.topaz

removeallmethods BtreeLeafNode
removeallclassmethods BtreeLeafNode
protectmethods

category: 'For Documentation Installation only'
classmethod: BtreeLeafNode
installDocumentation

< unprotected >

| doc txt |
doc := GsClassDocumentation _newForPrivateGsClass: self.

txt := (GsDocText new) details:
'A BtreeLeafNode represents an exterior node of a B-tree.' .
doc documentClassWith: txt.

self description: doc.
%

category: 'Comparison Operators'
method: BtreeLeafNode
_compareKey: aKey value: aValue equalToEntryAt: index 

"Performs a = comparison between aKey and the entry whose key is at the given
 index.  If the keys are equal, use the OOP of the value as the next basis for
 comparison.  The default implementation uses no encryption."
<protected>

^ (aKey _idxCompareEqualTo: (super at: index)) _and:
   [ aValue identityHash = (super at: (index - 1)) identityHash ]
%

category: 'Comparison Operators'
method: BtreeLeafNode
_compareKey: aKey value: aValue lessThanOrEqualToEntryAt: index

"Performs a <= comparison between aKey and the entry whose key is at the given
 index.  If the keys are equal, then use the OOP of the value for the
 comparison.  The default implementation uses no encryption."
<protected>

" first compare the keys "
( aKey _idxCompareLessThan: (super at: index) )
    ifTrue: [ ^ true ].

" if the keys are equal, use the OOP of the value "
(aKey _idxCompareEqualTo: (super at: index))
    ifTrue: [ ^ aValue identityHash <= (super at: (index - 1)) identityHash ].

^ false
%

category: 'Searching'
method: BtreeLeafNode
_findAllValuesForKey: aKey into: aCollection

"Finds all the values for the given key, placing them in the collection.  For
 leaf nodes, must first find the appropriate entry, then iterate through
 successive entries until we reach the end or a key does not match."
<protected>

| index eSize maxIndex |
" find first entry that might contain the key "
index := self _findCoveringIndexForKey: aKey totalOrder: false.
index == 0 ifTrue: [ ^ self ].

eSize := self entrySize.
maxIndex := self _lastIndex.

" now scan entries until reach the end or no match "
[ (index < maxIndex) _and:
[ self _compareKey: aKey equalToEntryAt: (index + 1) ] ] whileTrue: [
    aCollection add: (super at: index).
    index := index + eSize
]
%

category: 'Searching'
method: BtreeLeafNode
_findAllValuesForIdenticalKey: aKey into: aCollection

"Finds all the values for the given identical key, placing them in the
 collection.  For leaf nodes, must first find the appropriate entry, then
 iterate through successive entries until we reach the end or a key does
 not match.  Return whether the last key was equal or not (so the caller
 knows it must check the next leaf node)."
<protected>

| index eSize maxIndex |
" find first entry that might contain the key "
index := self _findCoveringIndexForKey: aKey totalOrder: false.
index == 0 ifTrue: [ ^ false ].

eSize := self entrySize.
maxIndex := self _lastIndex.

" now scan entries until reach the end or no match "
[ (index < maxIndex) _and:
[ self _compareKey: aKey equalToEntryAt: index + 1 ] ] whileTrue: [
  (aKey == (super at: index + 1))
    ifTrue: [ aCollection add: (super at: index) ].
  index := index + eSize
].
^ index >= maxIndex
%

category: 'Searching'
method: BtreeLeafNode
_findAllValuesFor: anObject identicalKey: aKey into: aCollection

"Finds all the values for the given identical key, placing only the objects that
 are identical to anObject in the collection.  For leaf nodes, must first find 
 the appropriate entry, then iterate through successive entries until we reach
 the end or a key does not match.  Return whether the last key was equal or not
 (so the caller knows it must check the next leaf node)."
<protected>

| index eSize maxIndex |
" find first entry that might contain the key "
index := self _findCoveringIndexForKey: aKey totalOrder: false.
index == 0 ifTrue: [ ^ false ].

eSize := self entrySize.
maxIndex := self _lastIndex.

" now scan entries until reach the end or no match "
[ (index < maxIndex) _and:
[ self _compareKey: aKey equalToEntryAt: index + 1 ] ] whileTrue: [
  (aKey == (super at: index + 1))
    ifTrue: 
      [ (super at: index) == anObject 
        ifTrue: [aCollection add: (super at: index)] ].
  index := index + eSize
].
^ index >= maxIndex
%

category: 'Searching Support'
method: BtreeLeafNode
_findAllValuesGreaterThanKey: aKey into: anArray

"Puts into anArray the index and the leaf node of the entry that is greater
 than the given key.  Returns true if a value was found, false otherwise."
<protected>

| index eSize lastIndex |
numElements == 0
    ifTrue: [ ^ false ].
lastIndex := self _lastKeyIndex.

" find index where the key would be inserted "
index := self _findCoveringIndexForKey: aKey totalOrder: false.
index == 0
    ifTrue: [ ^ false ].

eSize := self entrySize.
[ self _compareKey: aKey equalToEntryAt: index + 1 ] whileTrue: [
    index := index + eSize.
    index > lastIndex ifTrue: [ ^ false ]
].

anArray addLast: index.
anArray addLast: self.

^ true
%

category: 'Searching Support'
method: BtreeLeafNode
_findAllValuesLessThanKey: aKey andEquals: aBoolean into: anArray

"Puts into anArray the index and the leaf node of the entry that is less than
 (or less than and equal to if aBoolean is true) the given key.  Returns true
 if a value was found, false otherwise."
<protected>

| index eSize lastIndex |
numElements == 0
  ifTrue: [ ^ false ].

lastIndex := self _lastIndex.

" first see if all entries are < aKey "
(self _compareKey: aKey lessThanOrEqualToEntryAt: self _lastKeyIndex)
  ifFalse: [ " if so, put the rightmost leaf entry in the Array "
    self _putLastIndexOfLastChildInto: anArray.
    ^ true
  ].
" find index where the key would be inserted "
index := self _findCoveringIndexForKey: aKey totalOrder: false.

" if index == 1, check if any are < aKey "
index == 1
  ifTrue: [
    " if the first entry is greater than aKey, none satisfy "
    ( (self _compareKey: aKey lessThanEntryAt: 2) _or:
    [ aBoolean not _and:
    [ self _compareKey: aKey equalToEntryAt: 2 ] ] )
      ifTrue: [ ^ false ]
  ].

eSize := self entrySize.
" if also including exact matches "
aBoolean
  ifTrue: [ " scan forward until next entry that is greater "
    [ index < lastIndex _and:
    [ self _compareKey: aKey equalToEntryAt: index + 1 ] ]
      whileTrue: [ index := index + eSize ]
  ].
" subtract an entry to point to a valid 'less than' entry "
index > 1 ifTrue: [ index := index - eSize ].

anArray addLast: index.
anArray addLast: self.

^ true
%

category: 'Searching'
method: BtreeLeafNode
_findFirstValueForKey: aKey ifAbsent: aBlock

"Returns the first value associated with the given key.  For leaf nodes, find
 the appropriate entry and make the comparison."
<protected>

| index |

index := self _findCoveringIndexForKey: aKey totalOrder: false.
index == 0 ifTrue: [ ^ aBlock value ].

(self _compareKey: aKey equalToEntryAt: (index + 1))
    ifTrue: [ ^ super at: index ]
    ifFalse: [ ^ aBlock value ]
%

category: 'Searching'
method: BtreeLeafNode
_findFirstValueForKey: aKey

"Returns the first value associated with the given key.  For leaf nodes, find
 the appropriate entry and make the comparison.  If one is not found,
 returns the incomplete path traversal object."
<protected>

| index |

index := self _findCoveringIndexForKey: aKey totalOrder: false.
index == 0 ifTrue: [ ^ #_incompletePathTraversal ].

(self _compareKey: aKey equalToEntryAt: (index + 1))
    ifTrue: [ ^ super at: index ]
    ifFalse: [ ^ #_incompletePathTraversal ]
%

category: 'Updating'
method: BtreeLeafNode
_insertEncryptionFor: aKey value: aValue startingAt: insertionIndex

"Gets the encryption for the given key and place it in the receiver starting at
 the given index.  Subclasses may provide their own encryption; there is no
 default encryption."
<unprotected>

%

category: 'Searching Support'
method: BtreeLeafNode
_putFirstIndexOfFirstChildInto: anArray

"Puts the receiver and the first index of the receiver into the Array."
<protected>

numElements == 0
    ifTrue: [ anArray addLast: 0]
    ifFalse: [ anArray addLast: 1 ].
anArray addLast: self
%

category: 'Searching Support'
method: BtreeLeafNode
_putLastIndexOfLastChildInto: anArray

"Puts the receiver and the last index of the receiver into the Array."
<protected>

anArray addLast: self _lastKeyIndex - 1.
anArray addLast: self
%

category: 'Private'
method: BtreeLeafNode
_replaceKey: pretendKey with: realKey

"Due to a change to a byte object, an entry with a pretend key has been
 inserted into the B-tree.  Now replace all exact occurrences of the pretend
 key with the real key (which will be changed to match the pretend key later)."
<protected>

| index eSize lastIndex |

" faster to check result of binary search
(self _compareKey: pretendKey lessThanOrEqualToEntryAt: self _lastKeyIndex)
    ifFalse: [ ^ self ].
"

" find index where the key would be inserted "
index := self _binarySearchCoveringKey: pretendKey totalOrder: false.
" check if not found "
index > self _lastKeyIndex
    ifTrue: [ ^ self ].

eSize := self entrySize.
lastIndex := self _lastIndex.
" scan forward until no key matches "
[ index < lastIndex _and:
[ self _compareKey: pretendKey equalToEntryAt: index + 1 ] ]
    whileTrue: [
        " if key is exactly the pretendKey, replace it "
        ((super at: index + 1) == pretendKey)
            ifTrue: [ super at: index + 1 put: realKey ].

        index := index + eSize
    ]
%

category: 'Searching'
method: BtreeLeafNode
_scanForAllValuesForKey: aKey into: aCollection

"Places all values associated with the given key into aCollection.  Start with
 the first entry, and iterate through consecutive entries until we reach the
 end or the key does not match."
<protected>

| index eSize maxIndex |

numElements == 0 ifTrue: [ ^ self ].
index := 1.
eSize := self entrySize.
maxIndex := self _lastIndex.

" now scan entries until reach the end or no match "
[ (index < maxIndex) _and:
[ self _compareKey: aKey equalToEntryAt: (index + 1) ] ] whileTrue: [
    aCollection add: (super at: index).
    index := index + eSize
]
%

category: 'Updating'
method: BtreeLeafNode
_splitUsingKey: aKey value: aValue

"Creates a new sibling node, moving half of the receiver's entries into the
 sibling.  Then inserts a new entry with the given key/value into either the
 receiver or sibling.  Returns the new sibling."
<protected>

| index newSibling lastKeyIndex lastKey |
newSibling := self _createSibling.

" get the last key in the receiver "
lastKeyIndex := self _lastKeyIndex.
lastKey := super at: lastKeyIndex.

" determine if new entry goes in the first half or second half "
(self _compareKey: aKey value: aValue lessThanOrEqualToEntryAt: lastKeyIndex)
  ifTrue: [  " adding to the first half "
    index := self _binarySearchCoveringKey: aKey value: aValue.
    self _insertKey: aKey value: aValue atIndex: index
  ]
  ifFalse: [ " adding to the second half "
    index := newSibling _binarySearchCoveringKey: aKey value: aValue.
    newSibling _insertKey: aKey value: aValue atIndex: index
  ].

^ newSibling
%

category: 'Updating'
method: BtreeLeafNode
_updateLastValue

"For leaf nodes, the last value is derived, so nothing is required."
<unprotected>

%

category: 'Updating'
method: BtreeLeafNode
btreeAt: aKey put: aValue

""
<protected>

^ self at: aKey put: aValue
%

category: 'Updating'
method: BtreeLeafNode
at: aKey put: aValue

"Adds the key/value pair to the node.  If the node is full, performs a 'split'
 on the parent.  Returns the new sibling if a split is performed, otherwise
 returns the receiver."
<protected>

| index |
" determine if a split is needed "
numElements == self class maxNumberOfElements
  ifTrue: [ " performs the split and returns the new sibling "
    ^ self _splitUsingKey: aKey value: aValue
  ]
  ifFalse: [ " entry will fit in leaf node "
    numElements == 0
      ifTrue: [ index := 1 ]
      ifFalse: [
        index := self _binarySearchCoveringKey: aKey value: aValue.
      ].
    self _insertKey: aKey value: aValue atIndex: index
  ]
%

category: 'Testing'
method: BtreeLeafNode
isLeaf

"Returns whether the B-tree node is a leaf node."
<unprotected>

^ true
%

category: 'Accessing'
method: BtreeLeafNode
lastValue

"Returns the value for the last entry in the receiver."
<protected>

^ super at: (numElements - 1 * self entrySize + 1)
%

category: 'Constants'
method: BtreeLeafNode
mergeThreshold

"Returns the threshold such that if an entry has less than this number, it must
 have elements added from its sibling or be merged with its sibling.  This
 value is usually the same as the result of the message
 'self maxNumberOfElements quo: 2'."
<unprotected>

^ 500
%

category: 'Printing'
method: BtreeLeafNode
printOn: aStream withIndentationLevel: anInteger

"Puts a displayable representation of the keys of the receiver on the given
 stream, indenting the given number of spaces."
<protected>

anInteger timesRepeat: [ aStream nextPut: Character space ].

"aStream nextPutAll: '  { '; nextPutAll: self _alias asString; nextPutAll: ' } '."

2 to: (numElements * self entrySize) by: self entrySize do: [ :i |
    aStream nextPutAll: (super at: i) asString;
        nextPut: $(;
        nextPutAll: (super at: i - 1) asString;
        nextPutAll: ') '
].
aStream nextPut: Character lf
%

category: 'Printing'
method: BtreeLeafNode
printOn: aStream withIndentationLevel: anInteger includingValues: aBoolean

"Puts a displayable representation of the keys of the receiver on the given
 stream, indenting the given number of spaces."
<protected>

anInteger timesRepeat: [ aStream nextPut: Character space ].

"aStream nextPutAll: '  { '; nextPutAll: self _alias asString; nextPutAll: ' } '."

2 to: (numElements * self entrySize) by: self entrySize do: [ :i |
    aStream nextPutAll: (super at: i) asString.

    "aStream nextPut: $(;
        nextPutAll: (super at: i - 1) _alias asString;
        nextPut: $).
    "
    aBoolean
        ifTrue: [
            aStream nextPut: $(;
                nextPutAll: (super at: i - 1) asString;
                nextPutAll: ') '
        ]
        ifFalse: [ aStream nextPutAll: ' ' ]
].
aStream nextPut: Character lf
%

category: 'Removing'
method: BtreeLeafNode
removeKey: aKey value: aValue

"Removes the key/value pair.  Finds the index for the given key and then shifts
 any following entries to the left."
<protected>

| valueIndex eSize |
valueIndex := self _binarySearchForKey: aKey value: aValue.

valueIndex == 0 ifTrue: [ ^ false ].
eSize := self entrySize.

self _deleteNoShrinkFrom: valueIndex to: valueIndex + eSize - 1 .
numElements := numElements - 1.

^ true
%

category: 'Printing'
method: BtreeLeafNode
stringRep

""
<protected>

| aStream |

aStream := PrintStream printingOn: String new.
self printOn: aStream withIndentationLevel: 0.
^ aStream contents
%

category: 'Enumerating'
method: BtreeLeafNode
_preOrderDo: aBlock

"Execute the one-argument block with the receiver."
<protected>

aBlock value: self
%

category: 'Enumerating'
method: BtreeLeafNode
_leafKeysAndValuesDo: aBlock

"Execute the two-argument block with the key and value of each entry in the
 receiver."
<protected>

1 to: (numElements * self entrySize) by: self entrySize do: [ :i |
    aBlock value: (self _at: i + 1) value: (self _at: i)
]
%

category: 'Aggregate Support'
method: BtreeLeafNode
_numNonNilElements

"Returns the number of non-nil elements."
<protected>

| count |
count := 0.
" count from beginning the number of nils "
2 to: (numElements * self entrySize) by: self entrySize do: [ :i |
    nil == (self _at: i) ifTrue: [ count := count + 1]
].

^ numElements - count
%

category: 'Aggregate Support'
method: BtreeLeafNode
_putLeavesIn: array1 numNonNilsIn: array2

"Puts the receiver last in one Array and the number of non-nil elements in the
 other."
<protected>

array1 addLast: self.
array2 addLast: self _numNonNilElements
%

category: 'Private'
method: BtreeLeafNode
numElementsInStack: array
level: level
endNode: endNode
endIndex: endIndex
into: result

"Put the count in the first slot in the argument array and returns whether we
 reached the end."
<protected>

" if receiver is the beginning node "
self == (array at: level + 1)
    ifTrue: [
        " if receiver is also the ending node "
        self == endNode
            ifTrue: [ " all entries within this node "
                result at: 1 put: ( endIndex - (array at: level)) // self entrySize.
                ^ true
            ]
            ifFalse: [ " first elements are in this node "
                result at: 1 put: numElements - ((array at: level) // self entrySize).
                ^ false
            ]
    ]
    ifFalse: [ 
        " if receiver is the ending node "
        self == endNode
            ifTrue: [ " last elements are in this node "
                result at: 1 put: endIndex // self entrySize.
                ^ true
            ]
            ifFalse: [ " all elements are in this node "
                result at: 1 put: numElements.
                ^ false
            ]
    ]
%

category: 'Searching Support'
method: BtreeLeafNode
_lastValuesNotIdenticalTo: aKey into: array

""
<protected>

1 to: (numElements * self entrySize) by: self entrySize do: [ :i |
  ((self _at: i + 1) == aKey)
        ifFalse: [ array add: (self _at: i) ]
].
^ array
%

category: 'Sorting Support'
method: BtreeLeafNode
_putAscendingValuesInto: array startingAt: offset

"Returns the number of elements."
<protected>

| j |
j := offset.
1 to: (numElements * self entrySize) by: self entrySize do: [ :i |
  array at: j put: (self _at: i).
  j := j + 1
].
^ numElements
%

category: 'Sorting Support'
method: BtreeLeafNode
_putDescendingValuesInto: array startingAt: offset

"Returns the number of elements."
<protected>

| j |
j := offset.
1 to: (numElements * self entrySize) by: self entrySize do: [ :i |
  array at: j put: (self _at: i).
  j := j - 1
].
^ numElements
%

category: 'Accessing'
method: BtreeLeafNode
_totalElementsIn: stack endNode: endNode endIndex: endIndex into: array

"Determine the total number of entries, checking the given stack and
 end node from a stream object.  The given Array contains the current
 sum of elements in the first slot, and the second slots indicates if
 the end was hit."
<protected>

| i start end sum |
(i := stack indexOf: self) > 0
  ifTrue: [ start := stack at: i - 1 ]
  ifFalse: [ start := 1 ].

self == endNode
  ifTrue: [
    end := endIndex.
    array at: 2 put: true.
  ]
  ifFalse: [ end := self _lastIndex ].

sum := (array at: 1) + ((end - start) // self entrySize) + 1.
array at: 1 put: sum
%

category: 'Accessing'
method: BtreeLeafNode
totalElements

"Return the number of elements."
<protected>

^ numElements
%

category: 'Testing'
method: BtreeLeafNode
_isSymbolAt: offset

"Returns true if the key at the given offset is a Symbol; returns false
 otherwise."

<protected>

^ (self _basicAt: offset) _isSymbol
%

category: 'Searching'
method: BtreeLeafNode
_findFirstValueForKeyNoStrings: aKey

"Returns the first value associated with the given key.  For leaf nodes, finds
 the appropriate entry and makes the comparison.  If one is not found,
 returns the incomplete path traversal object."

<protected>

| index eSize maxIndex |

index := self _findCoveringIndexForKey: aKey totalOrder: false.
index == 0 ifTrue: [ ^ #_incompletePathTraversal ].

eSize := self entrySize.
maxIndex := self _lastIndex.

" now scan entries until reach the end or no match "
[ (index < maxIndex) _and:
[ self _compareKey: aKey equalToEntryAt: (index + 1) ] ] whileTrue: [
  (self _isSymbolAt: index + 1)
    ifTrue: [ ^ super at: index ].
  index := index + eSize
].
^ #_incompletePathTraversal
%

category: 'Searching'
method: BtreeLeafNode
_findFirstValueForKeyNoSymbols: aKey

"Returns the first value associated with the given key.  For leaf nodes, finds
 the appropriate entry and makes the comparison.  If one is not found,
 returns the incomplete path traversal object."

<protected>

| index eSize maxIndex |

index := self _findCoveringIndexForKey: aKey totalOrder: false.
index == 0 ifTrue: [ ^ #_incompletePathTraversal ].

eSize := self entrySize.
maxIndex := self _lastIndex.

" now scan entries until reach the end or no match "
[ (index < maxIndex) _and:
[ self _compareKey: aKey equalToEntryAt: (index + 1) ] ] whileTrue: [
  (self _isSymbolAt: index + 1)
    ifFalse: [ ^ super at: index ].
  index := index + eSize
].
^ #_incompletePathTraversal
%

unprotectmethods

category: 'Private'
method: BtreeLeafNode
_insertEncodingForString: aString encoding: numToEncode startingAt: startIdx

<primitive: 538>
aString  _validateClass: String .
numToEncode _validateClass: SmallInteger .
startIdx _validateClass: SmallInteger .

^ self _primitiveFailed: #_insertEncodingForString:encoding:startingAt:
%

