!=========================================================================
! Copyright (C) GemStone Systems, Inc. 1986-2011.  All Rights Reserved.
!
! $Id: rckeyvaluedict.gs,v 1.8 2008-01-09 19:02:36 stever Exp $
!
! Superclass Hierarchy:
!   RcKeyValueDictionary, AbstractDictionary, Collection, Object.
!
!=========================================================================

! class created in idxclasses.topaz

removeallmethods RcKeyValueDictionary
removeallclassmethods RcKeyValueDictionary

! ------------------- Class methods for RcKeyValueDictionary
category: 'Accessing the Class Format'
classmethod: RcKeyValueDictionary
firstPublicInstVar

"Returns the index of the first publicly available instance variable storage
 location, whether or not a public instance variable has actually been defined."

^ 3

%
category: 'Storing and Loading'
classmethod: RcKeyValueDictionary
loadFrom: passiveObj 

"Reads from passiveObj the passive form of an object with named instance
 variable format.  Converts the object to its active form by loading the
 information into a new instance of the receiver.  Returns the new instance."

| newDict oldDict size |
(passiveObj version >= 500) ifTrue:[ 
  size := passiveObj readSize.
  newDict := self new.
  newDict loadFrom: passiveObj size: size.
  ]
ifFalse:[
  "handle activation of objects written by 4.1.3"

  newDict := self new .
  passiveObj hasRead: newDict .
  size := passiveObj readSize. 
  oldDict := self new .
  oldDict basicLoadFromNoRead: passiveObj size: size .
  oldDict keysAndValuesDo:[:aKey :aValue | newDict at: aKey put: aValue ].
  oldDict initialize: 1 .
].
^ newDict

%
category: 'Instance Creation'
classmethod: RcKeyValueDictionary
new

"Returns a RcKeyValueDictionary with the default table size."

^ self new: 53

%
category: 'Instance Creation'
classmethod: RcKeyValueDictionary
new: tableSize

"Returns a RcKeyValueDictionary with the specified table size."

^ super _basicNew initialize: (Integer _selectedPrimeGreaterThan: tableSize)


%
! ------------------- Instance methods for RcKeyValueDictionary
category: 'Comparing'
method: RcKeyValueDictionary
= aKeyValueDictionary

"Compares two RcKeyValueDictionaries for equality.  Returns true if all of
 the following conditions are true:

 1.  the receiver and aKeyValueDictionary are of the same class,
 2.  the two RcKeyValueDictionaries have the same number of elements
 3.  all of the keys in one dictionary return the same value in
     both dictionaries.

 Returns false otherwise."

System _addRootObjectToRcReadSet: self.
( self class == aKeyValueDictionary class )
    ifFalse: [ ^ false ].

( self size == aKeyValueDictionary size )
    ifFalse: [ ^ false ].

System _addEntireObjectToRcReadSet: self.
self keysDo: [ :aKey |
    ( (self at: aKey)  == (aKeyValueDictionary at: aKey ifAbsent: [ ^ false ]))
        ifFalse: [ ^ false ]
].
^ true

%
category: 'Adding'
method: RcKeyValueDictionary
addAll: aCollection

"(R) Adds all the Association/key-value pairs contained in aCollection
 to the receiver.  Returns the argument, aCollection.

 The argument must be an AbstractDictionary or a collection of Associations."

"Note: In GemStone 4.1, this method returned the receiver."

(aCollection isKindOf: AbstractDictionary)
  ifTrue: [
    aCollection keysAndValuesDo: [ :aKey :aValue |
      self at: aKey put: aValue]
    ]
  ifFalse: [
    aCollection do: [ :anAssociation | self add: anAssociation ]
    ].

^ aCollection.

%
category: 'Enumerating'
method: RcKeyValueDictionary
associationsDo: aBlock

"(R) Iteratively evaluates the one argument block, aBlock, using an Association
 created by using each key-value pair successively.  Returns the receiver."

self keysAndValuesDo: [:aKey :aValue |
   aBlock value: (Association newWithKey: aKey value: aValue)
].

%
category: 'Accessing'
method: RcKeyValueDictionary
at: aKey ifAbsent: aBlock

"Returns the value that corresponds to aKey.  If no such key/value pair exists,
 returns the result of evaluating the zero-argument block aBlock."

| hash |
aKey == nil ifTrue: [ ^ self _error: #rtErrNilKey ].

hash := self hashFunction: aKey.
^ (self _rcAt: hash) at: aKey ifAbsent: aBlock

%
category: 'Accessing'
method: RcKeyValueDictionary
at: aKey otherwise: aValue

"Returns the value that corresponds to aKey.  If no such key/value pair exists,
 returns the given alternate value."

| hash |
aKey == nil ifTrue: [ ^ self _error: #rtErrNilKey ].

hash := self hashFunction: aKey.
^ (self _rcAt: hash) at: aKey otherwise: aValue

%
category: 'Updating'
method: RcKeyValueDictionary
at: aKey put: aValue

"(R) Adds the key-value pair to the receiver. If the receiver already contains 
 a key-value pair with the given key, this method makes aValue the value of that
 Association or key-value pair.  Returns aValue."

self _at: aKey put: aValue logging: true.
^ aValue

%
category: 'Private'
method: RcKeyValueDictionary
atHash: hashIndex put: aCollisionBucket

"Updates the hash table by storing aKey at the specified hashIndex.  This
 method is intended to be used internally.  If it is invoked directly by an
 application, concurrency conflicts may result (as well as destroying the
 consistency of the dictionary)."

((hashIndex <= 0) | (hashIndex > self _basicSize))
  ifTrue: [ hashIndex _error: #rtErrArgOutOfRange  .
            self _uncontinuableError
          ].
self basicAt: hashIndex put: aCollisionBucket

%
category: 'Storing and Loading'
method: RcKeyValueDictionary
basicWriteTo: passiveObj

"Converts the receiver to its passive form and writes that information on
 passiveObj."

| s cls ivs c |
  cls := self class.
  passiveObj writeClass: cls.

  ivs := cls _instVarNames .

  passiveObj writeSize: (s := self size) .

  cls firstPublicInstVar to: cls instSize do: [:i |
    (self shouldWriteInstVar: (ivs at: i)) ifTrue: [
      passiveObj writeObject: (self instVarAt: i) named: (ivs at: i)
    ].
  ].

  passiveObj endNamedInstVars.

  c := 0.
  self keysAndValuesDo:[ :aKey :aValue | 
    passiveObj writeObject: aKey; writeObject: aValue .
    c := c + 1.
    c > 49 ifTrue: [
      passiveObj lf.
      c := 0.
      ].
    ].

  passiveObj cr

%
category: 'Updating'
method: RcKeyValueDictionary
changeToSegment: segment

"Assigns the receiver and any collision buckets to the given segment."

self assignToSegment: segment.
self _doCollisionBuckets: [ :cb | cb assignToSegment: segment ]

%
category: 'Clustering'
method: RcKeyValueDictionary
clusterDepthFirst

"This method clusters the receiver and its named instance variables and the
 key/value pairs in depth-first order.  Returns true if the receiver has
 already been clustered during the current transaction; returns false
 otherwise."

System _addEntireObjectToRcReadSet: self.
self cluster
  ifTrue: [ ^ true ]
  ifFalse: [
    " cluster collision buckets "
    1 to: self _basicSize do: [ :i | (self basicAt: i) cluster ].

    " cluster instance variables "
    1 to: self class instSize do: [ :i | (self instVarAt: i) clusterDepthFirst ].

    " cluster keys and values "
    self keysAndValuesDo: [ :key :value |
      key clusterDepthFirst.
      value clusterDepthFirst
    ].
    ^ false
  ].

%
category: 'Copying'
method: RcKeyValueDictionary
copy

"Copies the collision buckets and returns a copy of the receiver."

| newKeyValueDict collisionBkt |
System _addEntireObjectToRcReadSet: self.
newKeyValueDict := super copy.
1 to: self _basicSize do: [ :index |
    collisionBkt := self basicAt: index.
    System _addEntireObjectToRcReadSet: collisionBkt.
    newKeyValueDict atHash: index put: collisionBkt copy
].
^ newKeyValueDict

%
category: 'Enumerating'
method: RcKeyValueDictionary
do: aBlock

"(R) Iteratively evaluates the one argument block aBlock, using the value
 of each Association as the argument of the block."
 
self keysAndValuesDo: [ :key :value | aBlock value: value ].

%
category: 'Comparing'
method: RcKeyValueDictionary
hash

"Since RcKeyValueDictionary | = is based on identity of elements, hash is based
 on identityHash of elements.  Returns the numerical hash value."

| hashValue |
System _addEntireObjectToRcReadSet: self.
hashValue := 97633.
self keysDo: [ :aKey |
    hashValue := hashValue bitXor: aKey hash
].
^ hashValue

%
category: 'Hashing'
method: RcKeyValueDictionary
hashFunction: aKey

"(R) The hash function should perform some operation on the value of the key 
 aKey which returns a value in the range 1..self _basicSize."

^ (aKey hash \\ self _basicSize) + 1

%
category: 'Initializing'
method: RcKeyValueDictionary
initialize: itsSize

"Initializes the instance variables of the receiver to be an empty
 RcKeyValueDictionary of the specified size.  This is intended to be used
 internally at the time of instance creation.  If used on an existing,
 populated RcKeyValueDictionary, concurrency conflicts may result."

"Returns the receiver."

| constraintClass sz theSize bucket |
theSize := itsSize .
(theSize <= 0)
    ifTrue: [ theSize _error: #rtErrArgNotPositive  .
              theSize := 1
    ].
collisionLimitPerBucket := theSize // 4.
constraintClass := self class varyingConstraint.
self _basicSize: theSize.
1 to: theSize do: [ :i |
    bucket := constraintClass new.
    bucket assignToSegment: self segment.
    bucket keyValueDictionary: self.
    self basicAt: i put: bucket.
].

sz := System rcValueCacheAt: #size for: self otherwise: nil.
sz ~~ nil
    ifTrue: [ System rcValueCacheAt: #size put: 0 for: self ]

%
category: 'Accessing'
method: RcKeyValueDictionary
keyAtValue: anObject ifAbsent: aBlock

"(R) Returns the key of the first value equal to the given object, anObject.
 If no match is found, evaluates and returns the result of the block aBlock."

"Note: In GemStone 4.1, this method checked for a value identical to anObject
 as opposed to checking for equality."

self keysAndValuesDo: [ :aKey :aValue |
  (anObject = aValue) ifTrue: [ ^ aKey ].
  ].
^ aBlock value.

%
category: 'Accessing'
method: RcKeyValueDictionary
keys

"Returns an IdentitySet containing the receiver's keys."

| keys |
keys := IdentitySet new.
self keysDo: [ :aKey | keys add: aKey ].
^ keys

%
category: 'Enumerating'
method: RcKeyValueDictionary
keysAndValuesDo: aBlock

"Iteratively evaluates the two argument block, aBlock, using each key and value
 of the receiver as the argument to the block. Returns the receiver."

System _addEntireObjectToRcReadSet: self.
self _doCollisionBuckets: [ :collisionBkt |
    1 to: collisionBkt tableSize do: [ :j | | aKey |
        aKey := collisionBkt keyAt: j.
        nil == aKey
            ifFalse: [ | aValue |
                aValue := collisionBkt valueAt: j.
                aBlock value: aKey value: aValue.
            ]
    ]
].


%
category: 'Enumerating'
method: RcKeyValueDictionary
keysDo: aBlock

"(R) Iteratively evaluates the two argument block aBlock, using each key and 
 value of the receiver as the argument to the block."

System _addEntireObjectToRcReadSet: self.
self _doCollisionBuckets: [ :collisionBkt | collisionBkt keysDo: aBlock ]

%
category: 'Storing and Loading'
method: RcKeyValueDictionary
loadVaryingFrom: passiveObj size: varyingSize

"Reads the varying part of the receiver from the given passive object.
 Does not record the receiver as having been read.  Does not read the
 receiver's named instance variables, if any."

| oldBkt newBkt |

passiveObj version >= 500 ifTrue:[
  1 to: varyingSize do: [:i |
    self at: passiveObj readObject"aKey" put: passiveObj readObject"aValue"
    ].
  ]
ifFalse:[
  1 to: varyingSize do: [:i |
    oldBkt := passiveObj readObject .
    (oldBkt isKindOf: RcCollisionBucket) 
       ifTrue:[ newBkt:= oldBkt ]
       ifFalse:[ 
         (oldBkt isKindOf: ObsoleteRcCollisionBucket) ifFalse:[
            self _halt:'invalid kind of RcCollisionBucket'.
            ].
         newBkt := RcCollisionBucket new .
         newBkt keyValueDictionary: self .
         oldBkt keysAndValuesDo:[:aKey :aValue | newBkt at: aKey put: aValue ] .
         oldBkt _removeAll .
         ].
    self _basicAt: i put: newBkt 
    ].
  ].

%
category: 'Accessing'
method: RcKeyValueDictionary
numElements

"Same as size."

^ self size

%
category: 'Hashing'
method: RcKeyValueDictionary
rebuildTable: newSize

"Rebuilds the hash table by saving the current state, initializing and changing
 the size of the table, and adding the key value pairs saved back to the hash
 dictionary.  This method is intended to be used internally.  If it is invoked
 directly by an application, concurrency conflicts may result."

| origSize constraintClass saveArray mySeg thisKey collBkt |
(newSize <= 0)
    ifTrue: [ newSize _error: #rtErrArgNotPositive  .  ^ self ].

origSize := self _basicSize.

mySeg := self segment.
constraintClass := self class varyingConstraint.
saveArray := Array new: origSize.

" save all old buckets in an Array and create new ones."
1 to: origSize do: [ :i | | oldBkt |
  oldBkt := self _at: i .
  saveArray at: i put: oldBkt .

  collBkt := constraintClass new.
  collBkt assignToSegment: mySeg.
  collBkt keyValueDictionary: self.
  self _at: i put: collBkt.
].

super _basicSize: newSize.

newSize > origSize
  ifTrue: [ " growing the RcKeyValueDictionary "
    origSize + 1 to: newSize do: [ :i | 
      collBkt := constraintClass new.
      collBkt assignToSegment: mySeg.
      collBkt keyValueDictionary: self.
      self _at: i put: collBkt.
    ]
  ].

collisionLimitPerBucket := newSize // 4.

1 to: origSize do: [ :i |
  collBkt := saveArray _at: i.
  1 to: collBkt _basicSize by: 2 do: [ :j |
    thisKey := collBkt _at: j.
      nil == thisKey
        ifFalse: [ self _at: thisKey put: (collBkt _at: j+1) logging: false ].
    ].
  collBkt _removeAll .  "dereference parent from old bucket"
  ].
saveArray size: 0.

%
category: 'Repository Conversion'
method: RcKeyValueDictionary
rehashForConversion

"Private. Rehashes the receiver because the hash values of some of its keys
 may have changed."

| newSize |

newSize := (Integer _selectedPrimeGreaterThan: (self size)) max: (self _basicSize).
self rebuildTable: newSize .
^ self.

%
category: 'Removing'
method: RcKeyValueDictionary
removeKey: aKey ifAbsent: aBlock

"(R) Removes the key/value pair with key aKey from the receiver and returns the
 value.  If no key/value pair is present with key aKey, evaluates the
 zero-argument block aBlock and returns the result of that evaluation."

| hash collisionBkt aValue sz |
hash := self hashFunction: aKey.
collisionBkt := self _rcAt: hash.

aValue := collisionBkt
    removeKey: aKey
    ifAbsent: [ ^ aBlock value ].

" log the removal in the redo log "
self _logRemoveKey: aKey inCollisionBucket: collisionBkt.

(sz := System rcValueCacheAt: #size for: self otherwise: nil) == nil
    ifTrue: [ ^ aValue ].
System rcValueCacheAt: #size put: (sz - 1) for: self.

^ aValue

%
category: 'Accessing'
method: RcKeyValueDictionary
size

"Returns the number of key/value pairs in the receiver."

| sz |
sz := System rcValueCacheAt: #size for: self otherwise: nil.
sz == nil
    ifTrue: [ sz := self _calculateSize ].
^ sz

%
category: 'Statistics'
method: RcKeyValueDictionary
statistics

"Returns a Dictionary containing statistics that can be useful in determining
 the performance of a key-value dictionary, including the following information:

 * TotalCollisionBuckets: The number of collision buckets required to
                          implement the key-value dictionary.

 * AvgPairsPerBucket:     The average number of key/value pairs in each bucket.

 * LargestBucket:         The bucket having the most key/value pairs.  This
                          bucket contains the most keys for which the hash
                          function did not provide a good distribution over the
                          range of values in the table.

 Since RcKeyValueDictionaries are implemented differently than
 KeyValueDictionaries, the statistics are calculated as if the implementation
 were the same as KeyValueDictionaries.  To be specific, RcKeyValueDictionaries
 are implemented with all collision buckets.  Therefore, the calculations don't
 count CollisionBuckets with a size of 1 as a collision."

| bktCount pairCount maxBkt maxBktSize result |
bktCount := 0.
pairCount :=0.
maxBktSize := 0.
System _addEntireObjectToRcReadSet: self.
self _doCollisionBuckets: [ :collisionBkt |
    System _addRootObjectToRcReadSet: collisionBkt.
    " if bucket size = 1, this would not be a collision in an ordinary key-value dictionary "
    collisionBkt size > 1
        ifTrue: [
            bktCount := bktCount + 1.
            pairCount := pairCount + collisionBkt size - 1.
            (collisionBkt size > maxBktSize)
                ifTrue: [
                    maxBktSize := collisionBkt size.
                    maxBkt := collisionBkt
                ]
        ]
].

result := SymbolDictionary new.
result at: #TotalCollisionBuckets put: bktCount.
(bktCount > 0)
    ifTrue: [ result at: #AvgPairsPerBucket put: (pairCount / bktCount) asFloat ]
    ifFalse: [ result at: #AvgPairsPerBucket put: 0].
result at: #LargestBucket put: maxBkt.
^ result

%
category: 'Private'
method: RcKeyValueDictionary
tableSize

"(R) Returns the size of hash table used for storing the entries."

^ self _basicSize

%
category: 'Initializing'
method: RcKeyValueDictionary
tableSize: aSize

"Sets the size of the receiver to a new value.  Each collision bucket is reset
 to contain no elements.  This method is intended to be used internally.  If it
 is invoked directly by an application, concurrency conflicts may result."

| origSize constraintClass mySeg collBkt |
(aSize <= 0)
    ifTrue: [ aSize _error: #rtErrArgNotPositive  .  ^ self ].

origSize := self _basicSize.
" reset each collision bucket to be empty "
1 to: origSize do: [ :i | (self basicAt: i) reset ].

super _basicSize: aSize.

aSize > origSize
    ifTrue: [ " growing the RcKeyValueDictionary "
        mySeg := self segment.
        constraintClass := self class varyingConstraint.
        (origSize + 1) to: aSize do: [ :i |
            collBkt := constraintClass new.
            collBkt assignToSegment: mySeg.
            collBkt keyValueDictionary: self.
            self basicAt: i put: collBkt.
        ].
    ].

collisionLimitPerBucket := aSize // 4.

%
category: 'Accessing'
method: RcKeyValueDictionary
values

"Returns an OrderedCollection containing the receiver's values."

| values |
values := OrderedCollection new.
System _addEntireObjectToRcReadSet: self.
self valuesDo: [ :aValue | values add: aValue ].
^ values

%
category: 'Enumerating'
method: RcKeyValueDictionary
valuesDo: aBlock

"(R) Iteratively evaluates the one argument block aBlock, using the value
 of each Association as the argument of the block."

self do: aBlock

%
category: 'Updating'
method: RcKeyValueDictionary
_at: aKey put: aValue logging: aBoolean

"Stores the aKey/aValue pair in the key-value dictionary.  Rebuilds the hash
 table if the addition caused the number of collisions to exceed the limit
 allowed.  The argument aBoolean determines if an entry is placed in the redo
 log for this operation."

| hash collisionBkt sz num oldValue |
aKey == nil ifTrue: [ ^ self _error: #rtErrNilKey ].
hash := self hashFunction: aKey.
collisionBkt := self _rcAt: hash.
num := collisionBkt numElements.
oldValue := collisionBkt at: aKey put: aValue.

aBoolean
    ifTrue: [
        self _logAt: aKey put: aValue oldValue: oldValue inCollisionBucket: collisionBkt.

        num < collisionBkt numElements
            ifTrue: [
                (sz := System rcValueCacheAt: #size for: self otherwise: nil) ~~ nil
                    ifTrue: [ System rcValueCacheAt: #size put: (sz + 1) for: self ]

            ]
    ].

(collisionBkt size > collisionLimitPerBucket)
  ifTrue: [
    self rebuildTable: (Integer _selectedPrimeGreaterThan: self _basicSize * 2)
  ].

^ true

%
category: 'Accessing'
method: RcKeyValueDictionary
_calculateSize

"Calculates the number of key/value pairs in the receiver and places it in the
 RC value cache.  Returns the total."

| num collisionBkt |
num := 0.
System _addEntireObjectToRcReadSet: self.
1 to: self _basicSize do: [ :i |
    collisionBkt := self basicAt: i.
    System _addRootObjectToRcReadSet: collisionBkt.
    num := num + collisionBkt size
].
System rcValueCacheAt: #size put: num for: self.
^ num

%
category: 'Updating'
method: RcKeyValueDictionary
_collisionLimitPerBucket: aLimit

"Sets the collisionLimitPerBucket for the receiver."

collisionLimitPerBucket := aLimit

%
category: 'Private'
method: RcKeyValueDictionary
_deferredGciUpdateWith: valueArray

"Private."

1 to: valueArray size by: 2 do:[:j |
  self at: (valueArray at: j) put: (valueArray at: j + 1)
  ].

%
category: 'Enumerating'
method: RcKeyValueDictionary
_doCollisionBuckets: aBlock

"Evaluates aBlock with each of the receiver's collision buckets as the
 argument.  The argument aBlock must be a one-argument block."

| collisionBkt |
1 to: self _basicSize do: [ :tableIndex |
    collisionBkt := self basicAt: tableIndex.
    aBlock value: collisionBkt
]

%
category: 'Private'
method: RcKeyValueDictionary
_gciInitialize

"Private."

| cache |

self _basicSize == 0
  ifTrue: [ self initialize: 53 ].
1 to: self _basicSize do: [ :i | (self basicAt: i) reset ].

" remove receiver from rc value cache "
( (cache := System _rcValueCache) ~~ nil _and:
[ (cache at: self otherwise: nil) ~~ nil ] )
  ifTrue: [ cache removeKey: self ].

%
category: 'Locking Support'
method: RcKeyValueDictionary
_lockableValues

"Returns an Array of the receiver's keys and values."

| result |
result := Array new .
self keysAndValuesDo: [:aKey :aValue |
  result add: aKey .
  result add: aValue .
] .
^ result

%
category: 'Private'
method: RcKeyValueDictionary
_logAt: aKey put: aValue oldValue: oldValue inCollisionBucket: aCollisionBkt

"Create a log entry for adding the given key and value.  The collision bucket
 is the object on which a conflict may occur."

| logEntry |
logEntry := LogEntry new.
logEntry receiver: self;
    selector: #_replayAt:put:oldValue:;
    argArray: #[ aKey, aValue, oldValue ].
System redoLog addLogEntry: logEntry forConflictObject: aCollisionBkt

%
category: 'Private'
method: RcKeyValueDictionary
_logRemoveKey: aKey inCollisionBucket: aCollisionBkt

"Create a log entry for removing the given key.  The collision bucket is the
 object on which a conflict may occur."

| logEntry |
logEntry := LogEntry new.
logEntry receiver: self;
    selector: #_removeKey:;
    argArray: (Array with: aKey).
System redoLog addLogEntry: logEntry forConflictObject: aCollisionBkt

%
category: 'Removing'
method: RcKeyValueDictionary
_removeKey: aKey

"Removes the key/value pair with key aKey from the receiver.  Returns true if
 the removal was successful, false if not.  This method is used when replaying
 operations from the redo log during conflict resolution."

| hash collisionBkt |
hash := self hashFunction: aKey.
collisionBkt := self _rcAt: hash.

collisionBkt removeKey: aKey ifAbsent: [  ^ false ].
^ true

%
category: 'Updating'
method: RcKeyValueDictionary
_replayAt: aKey put: aValue oldValue: oldValue

"Stores the aKey/aValue pair in the key-value dictionary.  Rebuilds the hash
 table if the addition caused the number of collisions to exceed the limit
 allowed.  If the existing value at aKey is not oldValue, then this method
 returns false."

| hash collisionBkt oldVal |
hash := self hashFunction: aKey.
collisionBkt := self _rcAt: hash.
oldVal := collisionBkt at: aKey put: aValue.

oldVal ~~ oldValue
    ifTrue: [ ^ false ].

(collisionBkt size > collisionLimitPerBucket)
  ifTrue: [
    self rebuildTable: (Integer _selectedPrimeGreaterThan: self _basicSize * 2)
  ].

^ true

%
category: 'Private'
method: RcKeyValueDictionary
_resolveRcConflictsWith: conflictObjects

"A logical write-write conflict has occurred on the receiver.  The objects that
 had the actual physical write-write conflicts are in the conflictObjects
 Array.  Selectively abort the receiver and then attempt to replay the
 operations maintained in the System redo log.  Returns true if all the
 operations could be replayed; otherwise returns false."

^ self _abortAndReplay: conflictObjects

%
category: 'Private'
method: RcKeyValueDictionary
_selectiveAbort

"Abort the collision buckets."

super _selectiveAbort.
1 to: self _basicSize do: [ :index | (self basicAt: index) _selectiveAbort ]

%
category: 'Private'
method: RcKeyValueDictionary
_validateRcConflictsWith: conflictObjects

"If the only conflict on this object is for the root, then we return true.
 Otherwise return false (to fail the transaction) if the redo log contains
 the selector _removeKey: since we cannot guarantee success if the transaction
 did a removal from the RcKeyValueDictionary."

" if only had a physical conflict on the root "
( conflictObjects size == 1 _and: [ (conflictObjects at: 1) == self ] )
  ifTrue: [ ^ true ].

(self _checkLogEntriesForNoSelector: #_replayAt:put:oldValue:)
  ifFalse: [ ^ false ].

^ self _checkLogEntriesForNoSelector: #_removeKey:

%
