!=========================================================================
! Copyright (C) GemStone Systems, Inc. 1986-2011.  All Rights Reserved.
!
! $Id: unorderedcoll.gs,v 1.13 2008-01-09 19:02:37 stever Exp $
! 
! UnorderedColl.gs
!
! Superclass Hierarchy:
!    UnorderedCollection, Collection, Object.
!
!=========================================================================

removeallmethods UnorderedCollection
removeallclassmethods UnorderedCollection

category: 'For Documentation Installation only'
classmethod: UnorderedCollection
installDocumentation

| doc txt |
doc := GsClassDocumentation newForClass: self.

txt := (GsDocText new) details:
'UnorderedCollection is an abstract class for collections of objects whose
 elements are not logically arranged in any particular order.  The elements are
 also not physically stored in any fixed order.  Any implied ordering at any
 given time is independent of the order in which the elements were added to the
 collection and cannot be relied upon to persist.

 The elements of unordered collections are all of the same kind.  Unless
 restricted further by a subclass, the kind of elements in unordered collections
 is Object.  That is, the class of each element must simply be some kind of
 Object.

 You cannot add nil to any kind of unordered collection.  Attempts to do so have
 no effect.

 UnorderedCollection provides for fast associative access of collection elements
 in searches by means of the use of indexes with selection blocks.
 UnorderedCollection creates each index for an individual instance, where
 specified, and maintains that index thereafter unless it is removed
 explicitly.

 Indexing is done on instance variables, not on values returned by messages.
 When an index path is used as an argument to a method, it is specified by a
 String that consists of instance variable names separated by periods (such as
 the String ''instvar1.instvar2.instvar3'').  The ith name in the String
 corresponds to the ith position in the path.  A path String may include up to
 16 names and is limited to a total of 1024 Characters.

 If aPathString is an empty path (that is, a zero-length String), the method
 operates upon the elements of the receiver itself rather than upon the instance
 variables of those elements.

 For more information about index structures and path expressions, see the
 GemStone Programming Guide.' .
doc documentClassWith: txt.

txt := (GsDocText new) details:
'For GemStone internal use.' .
doc documentInstVar: #_varyingSize with: txt.

txt := (GsDocText new) details:
'For GemStone internal use.' .
doc documentInstVar: #_numEntries with: txt.

txt := (GsDocText new) details:
'For GemStone internal use.' .
doc documentInstVar: #_indexedPaths with: txt.

txt := (GsDocText new) details:
'For GemStone internal use.' .
doc documentInstVar: #_levels with: txt.

self description: doc.
%

!==========================================================================
! Below this point: only method definitions in fileout order
!==========================================================================

! ------------------- Class methods for UnorderedCollection
category: 'Accessing the Class Format'
classmethod: UnorderedCollection
firstPublicInstVar

"Returns the index of the first user-visible instance variable defined in this
 class, regardless of whether or not this class actually has user-visible
 instance variables."

^ 5
%
category: 'Storing and Loading'
classmethod: UnorderedCollection
loadFrom: passiveObj

"Reads from passiveObj the passive form of an object.  Converts the object to
 its active form by loading the information into a new instance of the receiver.
 Returns the new instance."

| size inst |
size := passiveObj readSize.
inst := self new.
inst loadFrom: passiveObj size: size.
^inst
%
category: 'Private'
classmethod: UnorderedCollection
_asCollectionForSorting

"Used by the sorting subsystem."

^ self.
%
category: 'Private'
classmethod: UnorderedCollection
_disableRcIndexLogging

"System-wide disabling of logging for RC indexes
 (only for the life of the transaction)."

System rcValueCache at: #rcIndexLogging put: false
%
category: 'Private'
classmethod: UnorderedCollection
_enableRcIndexLogging

"System-wide enabling of logging for RC indexes."

System rcValueCache at: #rcIndexLogging put: true
%
category: 'Private'
classmethod: UnorderedCollection
_isRcIndexLoggingEnabled

""

^ System rcValueCacheAt: #rcIndexLogging otherwise: true.
%
! ------------------- Instance methods for UnorderedCollection
category: 'Adding'
method: UnorderedCollection
add: anObject withOccurrences: anInteger

"Includes anObject as an element of the receiver anInteger number of times.
 Generates an error if anObject is not a kind of the Bag's element kind."

self subclassResponsibility: #add:withOccurrences:
%
category: 'Storing and Loading'
method: UnorderedCollection
basicLoadFrom: passiveObj size: varyingSize

"Reads from passiveObj the passive form of an object.  Converts the object to
 its active form by loading the information into the receiver."

"This method is similar to basicLoadFrom:, but is used for objects whose size
 is not known when they are first instantiated (such as an IdentitySet)."

passiveObj hasRead: self.
^ self basicLoadFromNoRead: passiveObj size: varyingSize .
%
category: 'Storing and Loading'
method: UnorderedCollection
basicLoadFromNoRead: passiveObj size: varyingSize

"Private."

varyingSize == 0 ifTrue: [
  "Old NSC format had no named instance variable section.  A zero-length NSC in
   the old format might read instance variables from an enclosing object if
   there were not a special delimiter that could be reliably found."
  passiveObj checkForBagMark ifFalse: [
    passiveObj checkForInstVarMark ifFalse: [
      ^self
    ].
  ].
].
(passiveObj readNamedIV) ifFalse: [
  "old NSC format with no named instance variables"
  ^self loadVaryingFrom: passiveObj size: varyingSize
].
self loadNamedIVsFrom: passiveObj.
self loadVaryingFrom: passiveObj size: varyingSize.
%
category: 'Storing and Loading'
method: UnorderedCollection
basicWriteTo: passiveObj

"Converts the receiver to its passive form and writes that information on
 passiveObj."

| s cls ivs c |
  "store my structure on the given passiveObj"
  cls := self class.
  passiveObj writeClass: cls.

  ivs := cls _instVarNames.

  passiveObj writeSize: (s := self size) .

  cls firstPublicInstVar to: cls instSize do: [:i |
    (self shouldWriteInstVar: (ivs at: i)) ifTrue: [
      passiveObj writeObject: (self instVarAt: i) named: (ivs at: i)
    ].
  ].

  passiveObj writeBagMark.
  passiveObj endNamedInstVars.

  c := 0.
  self do: [ :x | 
    passiveObj writeObject: x.
    c := c + 1.
    c > 99 ifTrue: [
      passiveObj lf.
      c := 0.
    ].
  ].
  passiveObj cr
%
category: 'Clustering'
method: UnorderedCollection
clusterDepthFirst

"Clusters the receiver and its named and unnamed instance variables in
 depth-first order.  If indexes are present, the internal indexing objects are
 clustered also.  Returns true if the receiver has already been clustered during
 the current transaction; returns false otherwise."

self clusterIndexes.
^ super clusterDepthFirst
%
category: 'Clustering'
method: UnorderedCollection
clusterIndexes

"Clusters internal indexing objects using the current default ClusterBucket."

self _performNoDebug: #_clusterIndexes
%
category: 'Repository Conversion'
method: UnorderedCollection
convRemoveIndexes

"Private. Removes the indexes for the receiver. This method is to be used
 only during Repository Conversion."

<primitive: 901>
self _clearIndexList.
System _disableProtectedMode.
%
category: 'Copying'
method: UnorderedCollection
copy

"Returns a copy of the receiver that shares the receiver's public instance
 variables but has no indexes."

| result |
result := super copy .
result _finishShallowCopy .
^ result
%
category: 'Updating Indexes'
method: UnorderedCollection
createEqualityIndexOn: aPathString

"Creates an equality index on aPathString.  Generates an error if aPathString
 is not a path for the element kind of the receiver or if any term of the path
 is not constrained."

"If an error occurs during index creation, it may not be possible to commit the
 current transaction later."

"For index creation behaviour regarding transaction mode and transaction state,
 see comment in createEqualityIndexOn:commitInterval:"

^ self createEqualityIndexOn: aPathString commitInterval: nil
%
category: 'Updating Indexes'
method: UnorderedCollection
createEqualityIndexOn: aPathString commitInterval: anInteger

"Creates an equality index on aPathString.  Generates an error if aPathString
 is not a path for the element kind of the receiver or if any term of the path
 is not constrained.  While index creation is in progress, a transaction commit
 is performed after the given number of elements in the NSC have been processed.
 This method is intended to speed up index creation for very large collections."

"If an error occurs during index creation, it may not be possible to commit the
 current transaction later."

"Behavior of index creation with respect to transaction state:

When in manual mode:
    When using a commit interval:
        When originally outside of a transaction:
            Begin a transaction.
            Commit the final transaction of index creation.
            End outside of a transaction.
        When originally inside of a transaction:
            Commit the final transaction of index creation.
            End outside of a transaction.
    When not using a commit interval:
        When originally outside of a transaction:
            Signal an error.
      When originally inside of a transaction:
            End with the same transaction in progress
When in auto mode:
    When using a commit interval:
        Commit the final transaction of index creation.
    When not using a commit interval:
        End with the same transaction in progress

It is not appropriate to create indexes in transactionless mode.

"

<primitive: 901>

| result |

anInteger ~~ nil
  ifTrue: [
    anInteger <= 0
      ifTrue:[ anInteger _error: #rtErrArgNotPositive ].
  ].

self _checkIndexPathExpression: aPathString.
self segment currentUserCanWrite
  ifFalse: [ self _error: #authErrSegWrite args: #[ self segment ] ].

System transactionMode == #manualBegin ifTrue:
    [nil ~~ anInteger 
         ifTrue: 
             [System inTransaction ifFalse:
                  [System beginTransaction]]
         ifFalse: 
             [System inTransaction ifFalse:
                  [self _error: #rtErrPrimOutsideTrans]]].

Exception
  category: nil
  number: nil
  do: [ :ex :cat :num :args |
    ex remove.
    ex resignal: cat number: num args: args
    ].

result := self _performNoDebug: #_createEqualityIndexOn:commitInterval:
  with: aPathString
  with: anInteger.

nil ~~ anInteger ifTrue:
    [System commitTransaction ifFalse: 
         [self _errorCouldNotCommitDuringIndexCreation]].

System _disableProtectedMode.
result ~~ nil
  ifTrue: [ ^ self _error: (result at: 1) args: (result at: 2) ]
%
category: 'Updating Indexes'
method: UnorderedCollection
createEqualityIndexOn: aPathString withLastElementClass: aClass

"Creates an equality index on the path specified by aPathString.  The equality
 index will be ordered according to the sort provided comparison operators
 provided by aClass."

"Note: An error occurring during index creation may cause the
 current transaction to be unable to commit."

"For index creation behaviour regarding transaction mode and transaction state,
 see comment in createEqualityIndexOn:commitInterval:"

^ self createEqualityIndexOn: aPathString
  withLastElementClass: aClass
  commitInterval: nil
%
category: 'Updating Indexes'
method: UnorderedCollection
createEqualityIndexOn: aPathString
withLastElementClass: aClass
commitInterval: anInteger

"Creates an equality index on the path specified by aPathString.  The equality
 index is ordered according to the sort-provided comparison operators provided
 by aClass.   While index creation is in progress, a transaction commit is
 performed after the given number of elements in the NSC have been processed.
 This method is intended to speed up index creation for very large collections."

"Note: An error occurring during index creation may cause the
 current transaction to be unable to commit."

"For index creation behaviour regarding transaction mode and transaction state,
 see comment in createEqualityIndexOn:commitInterval:"

<primitive: 901>
| result |

"validate user parameters."
aClass _validateClass: Behavior.

anInteger ~~ nil
  ifTrue: [
    anInteger <= 0
      ifTrue:[ anInteger _error: #rtErrArgNotPositive ].
  ].

self _checkIndexPathExpression: aPathString.

self segment currentUserCanWrite
  ifFalse: [ self _error: #authErrSegWrite args: #[ self segment ] ].

System transactionMode == #manualBegin ifTrue:
    [nil ~~ anInteger 
         ifTrue: 
             [System inTransaction ifFalse:
                  [System beginTransaction]]
         ifFalse: 
             [System inTransaction ifFalse:
                  [self _error: #rtErrPrimOutsideTrans]]].

Exception
  category: nil
  number: nil
  do: [ :ex :cat :num :args |
    ex remove.
    ex resignal: cat number: num args: args
    ].

result := self
  _performNoDebug: #_createEqualityIndexOn:withLastElementClass:commitInterval:
  with: aPathString
  with: aClass
  with: anInteger.

nil ~~ anInteger ifTrue:
    [System commitTransaction ifFalse: 
         [self _errorCouldNotCommitDuringIndexCreation]].

System _disableProtectedMode.
result ~~ nil
  ifTrue: [ ^ self _error: (result at: 1) args: (result at: 2) ]
%
category: 'Updating Indexes'
method: UnorderedCollection
createIdentityIndexOn: aPathString

"Creates an identity index on aPathString.  Generates an error if aPathString
 is not a path for the element kind of the receiver or if any term of the path
 except the last term is not constrained."

"If an error occurs during index creation, it may not be possible to commit the
 current transaction later."

"For index creation behaviour regarding transaction mode and transaction state,
 see comment in createEqualityIndexOn:commitInterval:"

^ self createIdentityIndexOn: aPathString commitInterval: nil
%
category: 'Updating Indexes'
method: UnorderedCollection
createIdentityIndexOn: aPathString commitInterval: anInteger

"Creates an identity index on aPathString.  Generates an error if aPathString
 is not a path for the element kind of the receiver or if any term of the path
 except the last term is not constrained.  While index creation is in progress,
 a transaction commit is performed after the given number of elements in the NSC
 have been processed.  This method is intended to speed up index creation for
 very large collections."

"If an error occurs during index creation, it may not be possible to commit the
 current transaction later."

"For index creation behaviour regarding transaction mode and transaction state,
 see comment in createEqualityIndexOn:commitInterval:"

<primitive: 901>

anInteger ~~ nil
  ifTrue: [
    anInteger <= 0
      ifTrue:[ anInteger _error: #rtErrArgNotPositive ].
  ].

self _checkIndexPathExpression: aPathString.
self segment currentUserCanWrite
  ifFalse: [ self _error: #authErrSegWrite args: #[ self segment ] ].

System transactionMode == #manualBegin ifTrue:
    [nil ~~ anInteger 
         ifTrue: 
             [System inTransaction ifFalse:
                  [System beginTransaction]]
         ifFalse: 
             [System inTransaction ifFalse:
                  [self _error: #rtErrPrimOutsideTrans]]].

Exception
  category: nil
  number: nil
  do: [ :ex :cat :num :args | | txt |
    " get the text for the raised error "
    txt := cat textForError: num args: args.
    " check for recursive signal "
    num = (ErrorSymbols at: #rtErrPreventingCommit)
      ifTrue: [ " remove this exception and resignal "
        ex remove.
        anInteger ~~ nil
          ifTrue: [ self removeIncompleteIndex ].
        ex resignal: cat number: num args: args
      ]
      ifFalse: [ " append additional message to the end of text "
        txt _error: #rtErrPreventingCommit
      ]
  ].

self _performNoDebug: #_createIdentityIndexOn:commitInterval:
  with: aPathString
  with: anInteger.

nil ~~ anInteger ifTrue:
    [System commitTransaction ifFalse: 
         [self _errorCouldNotCommitDuringIndexCreation]].

System _disableProtectedMode.  "fix 18151"
%
category: 'Searching'
method: UnorderedCollection
detect: aBlock

"Evaluates aBlock repeatedly, with elements of the receiver as the argument.
 Returns the first element for which aBlock evaluates to true.  If none of the
 receiver's elements evaluates to true, generates an error.  The argument
 aBlock must be a one-argument block.  Uses associative access when the
 argument is a SelectionBlock."

<primitive: 901>
| result |
(aBlock _class == SelectBlock)
    ifTrue:[
        "send value: to the block to create the 4 element
        associative access Array; the argument nil is ignored to
        create the Array, but all SelectBlocks have one argument"
        result := (QueryExecuter on: self)
            _boundQueryDetect: (aBlock queryBlock value: nil).
        System _disableProtectedMode.
        result == #_incompletePathTraversal
            ifTrue: [
                ^ self _error: #assocErrNoElementsDetected args: #[aBlock]
            ].
        ^ result
       ].
System _disableProtectedMode.
^ self _detect: aBlock
%
category: 'Searching'
method: UnorderedCollection
_detect: aBlock

| each bag |
bag := self _asIdentityBag.
1 to: bag size do:[:i |
    each := (bag _at: i) .
    (aBlock value: each)
        ifTrue:[ ^ each ].
].
^ self _error: #assocErrNoElementsDetected args: #[aBlock] .
%
category: 'Searching'
method: UnorderedCollection
detect: aBlock ifNone: exceptionBlock

"Evaluates aBlock repeatedly, with elements of the receiver as the argument.
 Returns the first element for which aBlock has the value true.  If none of the
 receiver's elements has the value true, this evaluates the argument
 exceptionBlock and returns its value.  The argument aBlock must be a
 one-argument block, and exceptionBlock must be a zero-argument block.  Uses
 associative access when the argument is a SelectionBlock."

<primitive: 901>
| result |
(aBlock _class == SelectBlock)
    ifTrue:[
        "send value: to the block to create the 4 element
        associative access Array; the argument nil is ignored to
        create the Array, but all SelectBlocks have one argument"
        result := (QueryExecuter on: self)
            _boundQueryDetect: (aBlock queryBlock value: nil).
        System _disableProtectedMode.
        result == #_incompletePathTraversal
            ifTrue: [ ^ exceptionBlock value ].
        ^ result
    ].
System _disableProtectedMode.
^ self _detect: aBlock ifNone: exceptionBlock
%
category: 'Searching'
method: UnorderedCollection
_detect: aBlock ifNone: exceptionBlock

| each bag |
bag := self _asIdentityBag.
1 to: bag size do:[:i |
    each := (bag _at: i).
    (aBlock value: each)
        ifTrue:[ ^ each ].
].
^ exceptionBlock value

%
category: 'Accessing Indexes'
method: UnorderedCollection
equalityIndexedPaths

"Returns an Array of Strings, each of which represents a path for which an
 equality index exists in the receiver.  Each path originates with the elements
 of the receiver."

<primitive: 901>
| anArray |
anArray := Array new.
self _indexedPaths == nil
  ifTrue: [
    System _disableProtectedMode.
    ^ anArray
  ].
self _indexedPaths indexObjectsAndOffsetsDo: [ :indexObj :offset | | str |
  indexObj isRangeEqualityIndex
    ifTrue: [
      str := indexObj pathComponentsStringStartingAt: offset.
      indexObj isComplete
        ifFalse: [ str add: ' (incomplete)' ].
      (anArray includesValue: str)
        ifFalse: [ anArray addLast: str ]
    ]
].
System _disableProtectedMode.
^ anArray
%
category: 'Accessing Indexes'
method: UnorderedCollection
equalityIndexedPathsAndConstraints

"Returns an Array containing info about equality indexes.  The Array consists of
 String/Class pairs.  The string represents a path of the receiver's element
 kind for which an equality index exists in the receiver.  The class is the
 constraint on the last element in the path."

<primitive: 901>
| anArray |
anArray := Array new.
self _indexedPaths == nil
    ifTrue: [
        System _disableProtectedMode.
        ^ anArray
    ].
self _indexedPaths indexObjectsAndOffsetsDo: [ :indexObj :offset |
    indexObj isRangeEqualityIndex
        ifTrue: [
            anArray addLast: (indexObj pathComponentsStringStartingAt: offset).
            anArray addLast: indexObj lastElementClass
        ]
].
System _disableProtectedMode.
^ anArray
%
category: 'Repository Conversion'
method: UnorderedCollection
getIndexInfo

"Returns indexing information for the receiver."

<primitive: 901>

| iList indexObj path kind lastConstraint result |
  result := Array new.
  iList := self _indexedPaths.
  1 to: iList size by: 2 do: [ :i |
  indexObj := iList at: i.
  (iList at: i + 1) == 1
    ifTrue: [
      " path is first arg "
      path := indexObj pathComponentsString41.

      " kind is true if identity index, false if equality index "
      kind := indexObj isIdentityIndex.

      " for equality indexes, lastConstraint is second argument "
      kind
        ifFalse: [ 
          lastConstraint := indexObj lastElementClass41.  
          (lastConstraint == SmallInteger)
            ifTrue: [ lastConstraint := Integer ].
          ].

      result add: (Array with: self with: path with: kind with: lastConstraint).
      ].
    ].
  System _disableProtectedMode.
  ^ result.
%
category: 'Repository Conversion'
method: UnorderedCollection
getIndexInfoVersion: versionNum

"Returns indexing information for the receiver."

<primitive: 901>

| iList indexObj path kind lastConstraint result |
  result := Array new.
  iList := self _indexedPaths.
  1 to: iList size by: 2 do: [ :i |
  indexObj := iList at: i.
  (iList at: i + 1) == 1
    ifTrue: [
      " path is first arg "
      path := indexObj pathComponentsStringVersion: versionNum.

      " kind is true if identity index, false if equality index "
      kind := indexObj isIdentityIndex.

      " for equality indexes, lastConstraint is second argument "
      kind
        ifFalse: [
          lastConstraint := indexObj lastElementClassVersion: versionNum.
          (lastConstraint == SmallInteger)
            ifTrue: [ lastConstraint := Integer ].
          ].

      result add: (Array with: self with: path with: kind with: lastConstraint).
      ].
    ].
  System _disableProtectedMode.
  ^ result.
%
category: 'Indexing Support'
method: UnorderedCollection
getLastElementConstraintOnPath: aPathString

"Returns the class that is the last constraint class along the given path
 string.  If any of the classes that are traversed is not constrained on the
 given instance variable name, nil is returned."

| tmpClass pathArray |
pathArray := aPathString asArrayOfPathTerms.

" if path is empty string, then just use the constraint on unnamed variables "
(pathArray size == 1 _and: [ (pathArray at: 1) == #'' ] )
  ifTrue: [ ^ self class elementConstraint ].

tmpClass := self class elementConstraint.
pathArray do: [ :pathName |
  ( tmpClass isNsc _and: [ pathName == #*] )
    ifTrue: [ tmpClass := tmpClass elementConstraint ]
    ifFalse: [ tmpClass := (tmpClass _constraintOn: pathName) ].
  tmpClass == nil
    ifTrue: [ ^ nil ]
].
^ tmpClass
%
category: 'Accessing Indexes'
method: UnorderedCollection
identityIndexedPaths

"Returns an Array of Strings, each of which represents a path for which an
 identity index exists in the receiver.  Each path originates with the elements
 of the receiver."

<primitive: 901>
| anArray |
anArray := Array new.
self _indexedPaths == nil
  ifTrue: [
    System _disableProtectedMode.
    ^ anArray
  ].

self _indexedPaths indexObjectsAndOffsetsDo:[ :indexObj :offset | | pathString|
  indexObj isIdentityIndex
    ifTrue: [
      pathString := indexObj pathComponentsStringStartingAt: offset.
      (anArray includesValue: pathString)
        ifFalse: [ anArray addLast: pathString ]
    ]
    ifFalse: [
      (self _isRangeable: indexObj lastElementClass)
        ifTrue: [
          pathString := indexObj pathComponentsStringStartingAt: offset.
          indexObj isComplete
            ifFalse: [ pathString add: ' (incomplete)' ].
          (anArray includesValue: pathString)
            ifFalse: [ anArray addLast: pathString ]
        ]
    ]
].
System _disableProtectedMode.
^ anArray
%
category: 'Searching'
method: UnorderedCollection
includes: anObject

"Returns true if anObject is equal to one of the elements of the receiver. 
 Returns false otherwise."

"Note: In GemStone 4.1, this method returned true only if one of the elements of
 the receiver was identical to anObject. For functionality similar to that 
 provided by GemStone 4.1, use #includesIdentical:."

self subclassResponsibility: #includes:
%
category: 'Searching'
method: UnorderedCollection
includesIdentical: anObject

"Returns true if anObject is identical to one of the elements of the receiver. 
 Returns false otherwise."

self subclassResponsibility: #includesIdentical:
%
category: 'Searching'
method: UnorderedCollection
includesValue: anObject

"Returns true if the receiver contains an object of the same value as the
 argument, anObject.  Returns false otherwise.  (Compare with includes:,
 which is based on identity.)"

self subclassResponsibility: #includesValue:
%
category: 'Accessing'
method: UnorderedCollection
instVarAt: aSmallInteger

"If the receiver has a publicly accessible named instance variable at
 index aSmallInteger, this returns its value.  Generates an error if
 aSmallInteger is not a SmallInteger or is out of bounds, or if the
 receiver has no publicly accessible named instance variables."

(UnorderedCollection instSize >= aSmallInteger _and:
[ 1 <= aSmallInteger ])
  ifTrue:[ ^ nil ].

^ super instVarAt: aSmallInteger
%
category: 'Updating'
method: UnorderedCollection
instVarAt: anIndex put: aValue

"Stores the argument aValue in the instance variable indicated by anIndex and
 returns aValue. Generates an error if (1) anIndex is not a SmallInteger,
 (2) anIndex is out of bounds or (3) if the receiver has no publicly accessible
 named instance variables."

"Note: In GemStone 4.1, this method returned the receiver."

(UnorderedCollection instSize < anIndex) ifTrue:[
  ^ super instVarAt: anIndex put: aValue
].
^ self _errorIndexOutOfRange: anIndex
%
category: 'Auditing'
method: UnorderedCollection
invalidReferenceAudit
"For an explanation of this method, refer to the comments in the method
 UnorderedCollection>>invalidReferenceAuditWithRepair:"

  ^self invalidReferenceAuditWithRepair: false
%
category: 'Auditing'
method: UnorderedCollection
invalidReferenceAuditWithRepair: aBool

"Audit and optionally repair UnorderedCollections which incorrectly reference
 objects previously removed from the collection.  In rare cases in releases
 prior to 6.0.1, objects removed from a large UnorderedCollection may continue
 to be referenced even though the removal operation was successful and the
 object no longer appeared in the collection.

 If aBool is false, an audit is performed and the collection is not modified.
 If aBool is true, any invalid references in the collection are also repaired.

 Returns the number of unique objects which are incorrectly referenced by the
 receiver."

<primitive: 547>
  self _primitiveFailed: #invalidReferenceAuditWithRepair:.
%
category: 'Auditing'
method: UnorderedCollection
invalidReferenceRepair
"For an explanation of this method, refer to the comments in the method
 UnorderedCollection>>invalidReferenceAuditWithRepair:"

  ^self invalidReferenceAuditWithRepair: true
%
category: 'Accessing Indexes'
method: UnorderedCollection
kindsOfIndexOn: aPathString

"Returns a Symbol that indicates the kinds of indexes into the receiver that
 exist on aPathString: #identity, #equality, #equalityAndIdentity, or #none
 (either aPathString is not a path for the element kind of the receiver, or no
 indexes into the receiver exist on aPathString)."

<primitive: 901>
| pathList pathArray |

self _indexedPaths == nil
    ifTrue: [
        System _disableProtectedMode.
        ^ #none
    ].

pathList := #[].
pathArray := aPathString asArrayOfPathTerms.
self _indexedPaths indexObjectsAndOffsetsDo: [ :indexObj :offset |
    ( offset <= indexObj size _and:
    [ pathArray size > (indexObj size - offset) _and:
    [ indexObj hasIndexOnPath: pathArray startingAt: offset ] ] )
        ifTrue: [ pathList add: indexObj ].
].

pathList size == 2
    ifTrue: [
        System _disableProtectedMode.
        ^ #equalityAndIdentity
    ].

pathList size == 1
    ifTrue: [
        pathList first isRangeEqualityIndex
            ifTrue: [
                ( self _isRangeable: pathList first lastElementClass )
                    ifTrue: [
                        System _disableProtectedMode.
                        ^ #equalityAndIdentity
                    ]
                    ifFalse: [
                        System _disableProtectedMode.
                        ^ #equality
                    ]
            ]
            ifFalse: [
                System _disableProtectedMode.
                ^ #identity
            ]
    ].

self _indexedPaths ~~ nil
    ifTrue: [
        (self _indexedPaths _numberOfCommonPathTermsForPathArray: pathArray) = pathArray size
            ifTrue: [
                System _disableProtectedMode.
                ^ #identity
            ]
            ifFalse: [
                System _disableProtectedMode.
                ^ #none
            ]
    ].

System _disableProtectedMode.
^ #none
%
category: 'Storing and Loading'
method: UnorderedCollection
loadFrom: passiveObj size: varyingSize

"Reads from passiveObj the passive form of an object.  Converts the object to
 its active form by loading the information into the receiver."

"This method is similar to loadFrom:, but is used for objects whose size
 is not known when they are first instantiated (such as an IdentitySet)."

^self basicLoadFrom: passiveObj size: varyingSize
%
category: 'Storing and Loading'
method: UnorderedCollection
loadNamedIVsFrom: passiveObj

"Reads named instance variables from the given passive object.  The first
 instance variable should already have been parsed and be available in the
 passiveObj argument."

| name offset nameSym |

[ name := passiveObj ivName.
  name ~~ nil ifTrue: [
    nameSym := Symbol _existingWithAll: name .
    nameSym ~~ nil ifTrue:[
      offset := self class _ivOffsetOf: nameSym.
      offset ~~ nil ifTrue:[ self instVarAt: offset put: passiveObj ivValue ]
             ifFalse:[ self obsoleteInstVar: nameSym value: passiveObj ivValue].
    ]
    ifFalse:[
      self obsoleteInstVar: name value: passiveObj ivValue
    ].
    passiveObj readNamedIV
  ]
  ifFalse: [
    false
  ]
] untilFalse.

passiveObj checkForBagMark.
passiveObj skipNamedInstVars.
%
category: 'Storing and Loading'
method: UnorderedCollection
loadVaryingFrom: passiveObj size: varyingSize

"Reads the varying part of the receiver from the given passive object.
 Does not record the receiver as having been read.  Does not read the
 receiver's named instance variables, if any."

1 to: varyingSize do: [:i |
  self add: (passiveObj readObject)
].
%
category: 'Instance Migration'
method: UnorderedCollection
migrateFrom: anotherObject instVarMap: otherivi

"(R) Takes information from the given object and puts it in the receiver.  This
 message is sent to an object when its class is being migrated to another class
 to account for changes in a schema.  The otherivi argument is a precalculated
 indirection table associating the receiver's instance variables with instance
 variables in the other object.  If a table entry is 0, the other object is
 assumed not to have that instance variable.

 This method should be augmented to perform other necessary initializations in
 the receiver."

| otherClass varyingConstraint |

super migrateFrom: anotherObject instVarMap: otherivi.

"if no elements in anotherObject, no need to check varying constraint "
anotherObject size == 0
  ifTrue: [ ^ self ].

otherClass := anotherObject class.
varyingConstraint := self class varyingConstraint.

otherClass isNonByteVarying
  ifTrue: [
    (otherClass varyingConstraint isSubclassOf: varyingConstraint)
      ifTrue: [ " no need to perform constraint checking "
        anotherObject do: [ :anElement | self add: anElement ].
        ]
      ifFalse: [ "otherwise check varying constraint"
        anotherObject do: [ :anElement | | newElement |
          " see if self can contain anElement "
          (anElement isKindOf: varyingConstraint)
            ifTrue: [ self add: anElement ]
            ifFalse: [
              "if not, see if anElement can be converted (or an error raised)"
              newElement :=
                anotherObject invalidElementConstraintWhenMigratingInto: self
                  for: anElement.
              "if anElement was converted, add it to the receiver"
              newElement ~~ anElement ifTrue: [ self add: newElement ].
              ]
          ]
        ]
    ].

^ self.
%
category: 'Searching'
method: UnorderedCollection
occurrencesOf: anObject

"Returns the number of the receiver's elements that are identical (==) to
 anObject."

self subclassResponsibility: #occurrencesOf:
%
category: 'Updating Indexes'
method: UnorderedCollection
progressOfIndexCreation

"Returns a String that describes the progress of an index creation that is
 underway."

<primitive: 901>
| indexObj cnt str lf |

indexObj := self _findIncompleteIndex.
( indexObj == nil _or:
" get the current offset of enumeration through the NSC "
[ (cnt := indexObj progress) == nil ] )
  ifTrue: [ 
    System _disableProtectedMode.
    ^ 'No index creation is in progress.' 
  ].

str := String new.
lf := Character lf.
str add: lf; add: 'Creating index on <';
  add: indexObj pathComponentsString; add: $>; add: lf;
  add: 'During enumeration of collection, finished ';
  add: cnt asString; add: ' out of ';
  add: self size asString; add: ' total elements.'.

indexObj isRangeEqualityIndex
  ifTrue: [
    " get the current offset of enumeration through the n-way merge "
    cnt := indexObj nwayMergeProgress.
    cnt ~~ nil
      ifTrue: [
        str add: lf; add: 'During n-way merge, finished ';
          add: cnt asString; add: ' elements'; add: lf.
      ]
  ].

System _disableProtectedMode.
^ str
%
category: 'Repository Conversion'
method: UnorderedCollection
rebuildIndexes: indexInfo

"Rebuilds the indexes."

| path kind lastConstraint newLastConstraint |

  path := indexInfo at: 2.
  kind := indexInfo at: 3.
  lastConstraint := indexInfo at: 4.

  kind
    ifTrue: [
      self createIdentityIndexOn: path
      ]
    ifFalse: [
      newLastConstraint := lastConstraint _correspondingNewClass.
      self createEqualityIndexOn: path withLastElementClass: newLastConstraint
      ].
%
category: 'Updating Indexes'
method: UnorderedCollection
recomputeIndexSegments

"Clears the segments set of each path term for all indexes, then traverses all
 elements in the receiver, adding the segment into the corresponding path
 term's segment set.  This method will have a high probability of causing
 concurrency conflicts with other sessions that modify the receiver."

<primitive: 901>
self _indexedPaths == nil
  ifTrue: [
    System _disableProtectedMode.
    ^ self
  ].

Exception
    category: nil
    number: nil
    do: [ :ex :cat :num :args | | txt |
        " get the text for the raised error "
        txt := cat textForError: num args: args.
        " check for recursive signal "
        num = (ErrorSymbols at: #rtErrPreventingCommit)
            ifTrue: [ " remove this exception and resignal "
                ex remove.
                ex resignal: cat number: num args: args
            ]
            ifFalse: [ " append additional message to the end of text "
                txt _error: #rtErrPreventingCommit
            ]
    ].

self _performNoDebug: #_recomputeIndexSegments.
System _disableProtectedMode
%
category: 'Searching'
method: UnorderedCollection
reject: aBlock

"Evaluates aBlock with each of the receiver's elements as the argument.
 Stores the values for which aBlock is false into a collection of the same
 class as the receiver, and returns the new collection.  The argument aBlock
 must be a one-argument block.  Uses associative access when the argument is
 SelectionBlock."

<primitive: 901>
| result |
(aBlock _class == SelectBlock)
    ifTrue:[
       "send value: to the block to create the 4 element
        associative access Array; the argument nil is ignored to
        create the Array, but all SelectBlocks have one argument"
        result := (QueryExecuter on: self)
            _boundQueryReject: (aBlock queryBlock value: nil).
        System _disableProtectedMode.
        ^ result
    ].
System _disableProtectedMode.
^ self _reject: aBlock
%
category: 'Searching'
method: UnorderedCollection
_reject: aBlock

| each bag result |
bag := self _asIdentityBag.
result:= NscBuilder for: self speciesForSelect new max: bag size.
1 to: bag size do:[:i |
    each := (bag _at: i).
    (aBlock value: each)
        ifFalse:[ result add: each ].
].
^ result completeBag
%
category: 'Updating Indexes'
method: UnorderedCollection
removeAllIndexes

"Remove all indexes for the receiver.  If the receiver contains implicit indexes
 (due to its participation as a set-valued instance variable in another NSC's
 index), this method returns an Array of pairs.  The first object in each pair
 is a root NSC that has an index, and the second object in the pair is a path
 string that causes the receiver to participate in an index.  If all of the
 receiver's indexes can be removed, this method returns the receiver."

"If an error occurs during index removal, it may not be possible to commit the
 current transaction later."

<primitive: 901>
| result iList rootIndexes hasNonRootIndex |
(iList := self _indexedPaths) == nil
    ifTrue: [
        System _disableProtectedMode.
        ^ self
    ].

hasNonRootIndex := false.
rootIndexes := Array new.
" scan each entry to see if a non-root index is present "
1 to: iList size by: 2 do: [ :i |
    " if it is a root index ... "
    (iList at: i + 1) == 1
        " add it to a list "
        ifTrue: [ rootIndexes add: (iList at: i) ]
        " otherwise flag that we're really not removing ALL indexes "
        ifFalse: [ hasNonRootIndex := true ]
].

" if all indexes are explicit, check authorization to remove them "
hasNonRootIndex
  ifFalse: [
    self segment currentUserCanWrite
      ifFalse: [ self _error: #authErrSegWrite args: #[ self segment ] ].
    iList _checkAllIndexAuthorizations.
  ].

Exception
    category: nil
    number: nil
    do: [ :ex :cat :num :args | | txt |
        " get the text for the raised error "
        txt := cat textForError: num args: args.
        " check for recursive signal "
        num = (ErrorSymbols at: #rtErrPreventingCommit)
            ifTrue: [ " remove this exception and resignal "
                ex remove.
                ex resignal: cat number: num args: args
            ]
            ifFalse: [ " append additional message to the end of text "
                txt _error: #rtErrPreventingCommit
            ]
    ].

result := self
  _performNoDebug: #_removeAllRootIndexes:hasNonRootIndex:
  with: rootIndexes
  with: hasNonRootIndex.

System _disableProtectedMode.
^ result
%
category: 'Removing'
method: UnorderedCollection
removeAllPresent: aCollection

"Removes from the receiver one occurrence of each element of aCollection that is
 also an element of the receiver.  Differs from removeAll: in that, if some
 elements of aCollection are not present in the receiver, no error is generated.
 Returns aCollection."

aCollection do: [ :anElement |
  self removeIfPresent: anElement
  ].

^ aCollection
%
category: 'Updating Indexes'
method: UnorderedCollection
removeEqualityIndexOn: aPathString

"If an equality index exists on aPathString, remove that index.  If the
 path string is invalid or no index exists on the given path, an error
 is raised.  If aPathString is an implicit index (due to the receiver's
 participation as a set-valued instance variable in some other NSC's
 index), then this method returns the path string."

"If an error occurs during index removal, it may not be possible to commit the
 current transaction later."

<primitive: 901>
| result |
self _checkIndexPathExpression: aPathString.

self _indexedPaths == nil
    ifTrue: [ self _error: #rtErrNoIndexForPath args: #[ aPathString ] ].

self segment currentUserCanWrite
  ifFalse: [ self _error: #authErrSegWrite args: #[ self segment ] ].

result := self _performNoDebug: #_removeEqualityIndexOn: with: aPathString.
result == nil
  ifTrue: [ self _error: #rtErrNoIndexForPath args: #[ aPathString ] ].

System _disableProtectedMode.
^ result
%
category: 'Updating Indexes'
method: UnorderedCollection
removeIdentityIndexOn: aPathString

"If an identity index exists on aPathString, and aPathString is not a
 proper prefix of some indexed path, the the index is removed.  If the
 path string is invalid or no index exists on the given path, an error
 is raised.  If aPathString is an implicit index (due to the receiver's
 participation as a set-valued instance variable in some other NSC's
 index), then this method returns the path string."

"If an error occurs during index removal, it may not be possible to commit the
 current transaction later."

<primitive: 901>
| result |
self _checkIndexPathExpression: aPathString.

self _indexedPaths == nil
    ifTrue: [ self _error: #rtErrNoIndexForPath args: #[ aPathString ] ].

self segment currentUserCanWrite
  ifFalse: [ self _error: #authErrSegWrite args: #[ self segment ] ].

result := self _performNoDebug: #_removeIdentityIndexOn: with: aPathString.
result == nil
  ifTrue: [ self _error: #rtErrNoIndexForPath args: #[ aPathString ] ].

System _disableProtectedMode.
^ result
%
category: 'Removing'
method: UnorderedCollection
removeIfPresent: anObject

"Removes anObject from the receiver and returns anObject.  If anObject is
 present several times in the receiver, only one occurrence is removed.
 Returns nil if anObject is missing from the receiver."

self remove: anObject ifAbsent: [ ^ nil ].
^ anObject.
%
category: 'Updating Indexes'
method: UnorderedCollection
removeIncompleteIndex

"If there is an incomplete index, clean it up."

<primitive: 901>
| indexObj |

indexObj := self _findIncompleteIndex.
indexObj == nil
  ifTrue: [ 
    System _disableProtectedMode.
    ^ self 
  ].

self _undoIndexCreation: indexObj pathTerm: indexObj _findFirstUnsharedPathTerm.
System _disableProtectedMode.
%
category: 'Searching'
method: UnorderedCollection
select: aBlock

"Evaluates aBlock with each of the receiver's elements as the argument.
 Stores the values for which aBlock is true into a collection of the same
 class as the receiver, and returns the new collection.  The argument aBlock
 must be a one-argument block.  Uses associative access when the argument is a
 SelectionBlock.

 The new collection returned by this method will not retain any indexes of
 the receiver.  If you want to perform indexed selections on the new
 collection, you must build all of the necessary indexes.  The discussion of
 'Transferring Indexes' in the 'Indexed Associative Access' chapter of the
 GemStone Programming Guide describes a technique for doing this."

<primitive: 901>
| result |
(aBlock _class == SelectBlock)
    ifTrue:[
       "send value: to the block to create the 4 element
        associative access Array; the argument nil is ignored to
        create the Array, but all SelectBlocks have one argument"
        result := (QueryExecuter on: self)
            _boundQuerySelect: (aBlock queryBlock value: nil).
        System _disableProtectedMode.
        ^ result
    ].
System _disableProtectedMode.
^ self _select: aBlock
%
category: 'Searching'
method: UnorderedCollection
_select: aBlock

| each bag result |
bag := self _asIdentityBag.
result:= NscBuilder for: self speciesForSelect new max: bag size.
1 to: bag size do:[:i |
    each := (bag _at: i) .
    (aBlock value: each)
        ifTrue:[ result add: each ].
].
^result completeBag
%
category: 'Searching'
method: UnorderedCollection
selectAsStream: aBlock

"Same functionality as select: except that the result is returned as a
 RangeIndexReadStream rather than an IdentitySet.  The select block is limited
 in the following ways:

 * The select block may only contain a single predicate.
 * The predicate must contain one path expression.
 * An equality index must exist for the path expression.

 To use the stream that this method returns most effectively, avoid modifying
 both the receiver of this message and the selected objects returned by the
 stream as long as the stream is being accessed.  Changes that alter the
 equality index can cause stream access failures."

<primitive: 901>
| result |
(aBlock _class == SelectBlock)
    ifTrue:[
       "send value: to the block to create the 4 element
        associative access Array; the argument nil is ignored to
        create the Array, but all SelectBlocks have one argument"
        result := (QueryExecuter on: self)
            _boundQuerySelectAsStream: (aBlock queryBlock value: nil).
        System _disableProtectedMode.
        ^ result
    ].
^ self _error: #rtErrBagOnlySelectBlockAllowed
%
category: 'Updating Indexes - Private'
method: UnorderedCollection
_addMappingsFor: indexObj
updateIndexListSegments: aBoolean
commitInterval: anInteger

"Adds the mappings in the index dictionary, dependency lists, and so on for all
 elements in the receiver."

< protected >

| pathTerm hasSet setOffset object objNsc bag updateIndexList iList |

" since this index may use path terms from other indexes,
find the first one that is not shared by another index (only
need to add mappings from that point onward) "
setOffset := indexObj _findSetValuedPathTermOffset.
iList := self _indexedPaths.

pathTerm := indexObj _findFirstUnsharedPathTerm.
pathTerm == nil
  ifTrue: [ " index is subsumed by an existing index "
    " see if there is a set-valued term "
    setOffset == 0
      ifFalse: [ " need to update index lists "
        bag := self _asIdentityBag.
        pathTerm := indexObj lastPathTerm.
        " if last path term is #*, will need to update the
          last objects' index list "
        pathTerm indicatesNsc
          ifTrue: [ objNsc := NscBuilder for: IdentityBag new max: indexObj sizeForNscBuilder ]
          ifFalse: [ objNsc := nil ].

        " if index list's segments set needs updated ... "
        aBoolean
          ifTrue: [
            1 to: bag size do: [ :i |
              object := bag _at: i.
              indexObj
                traverse: object
                upTo: pathTerm
                startingAt: 1
                into: objNsc
                addingIndex: true.
              iList addSegment: object segment logging: false

            ]
          ]
          ifFalse: [
            1 to: bag size do: [ :i |
              indexObj
                traverse: (bag _at: i)
                upTo: pathTerm
                startingAt: 1
                into: objNsc
                addingIndex: true
            ]
          ].

        objNsc ~~ nil
          ifTrue: [
            pathTerm addIndexListEntriesFor: objNsc completeBag index: indexObj
          ]
      ].
    ^ self
  ].

"Determine if there is a set-valued instance variable before the unshared path
 term."
setOffset := indexObj _findSetValuedPathTermOffset.
setOffset == 0
  ifTrue: [ hasSet := false ]
  ifFalse: [ hasSet := setOffset < pathTerm offset ].

bag := self _asIdentityBag.
hasSet
  ifTrue: [
    " if there is a set-valued instance variable, collect objects in a
      Bag on which to update the index dictionary "

    updateIndexList := false.
    " see if index is subsumed by an existing index "
    ( pathTerm == indexObj lastPathTerm _and:
    [ pathTerm size > 1 _and:
    [ (pathTerm at: 1) size ~= (pathTerm at: 2) size ] ] )
      ifTrue: [
        " if last path term is #*, will need to update the
          last objects' index list "
        pathTerm indicatesNsc
          ifTrue: [
            updateIndexList := true.
            objNsc := NscBuilder for: IdentityBag new max: indexObj sizeForNscBuilder
          ]
          ifFalse: [ objNsc := nil ]
      ]
      ifFalse: [ objNsc := NscBuilder for: IdentityBag new max: indexObj sizeForNscBuilder ].

    " if index list's segments set needs updated ... "
    aBoolean
      ifTrue: [
        1 to: bag size do: [ :i |
          object := bag _at: i.
          " traverse up to the point of the unshared path term "
          indexObj
            traverse: object
            upTo: pathTerm
            startingAt: 1
            into: objNsc
            addingIndex: true.
          iList addSegment: object segment logging: false
        ]
      ]
      ifFalse: [
        1 to: bag size do: [ :i |
          " traverse up to the point of the unshared path term "
          indexObj
            traverse: (bag _at: i)
            upTo: pathTerm 
            startingAt: 1
            into: objNsc
            addingIndex: true
        ]
      ].

    objNsc ~~ nil
      ifTrue: [
        updateIndexList
          ifTrue: [
            pathTerm addIndexListEntriesFor: objNsc completeBag index: indexObj
          ]
          ifFalse: [
            objNsc := objNsc completeBag.
            1 to: objNsc size do: [ :j |
              pathTerm addMappingsForObject: (objNsc _at: j) logging: false
            ]
          ]
      ]
  ]
  ifFalse: [
    anInteger ~~ nil
      ifTrue: [ self _lockForIndexCreation ].

    " see if need to traverse to the first unshared path term "
    indexObj firstPathTerm == pathTerm
      ifTrue: [
        " if index list's segments set needs updated ... "
        aBoolean
          ifTrue: [
            1 to: bag size do: [ :i |
              anInteger ~~ nil
                ifTrue: [
                  self _commitIndexCreation: indexObj
                    at: i
                    commitInterval: anInteger
                ].
              object := bag _at: i.
              pathTerm addMappingsForObject: object logging: false.
              iList addSegment: object segment logging: false
            ]
          ]
          ifFalse: [
            1 to: bag size do: [ :i |
              anInteger ~~ nil
                ifTrue: [
                  self _commitIndexCreation: indexObj
                    at: i
                    commitInterval: anInteger
                ].
              pathTerm addMappingsForObject: (bag _at: i) logging: false
            ]
          ]
      ]
      ifFalse: [
        " if index list's segments set needs updated ... "
        aBoolean
          ifTrue: [
            1 to: bag size do: [ :i |
              anInteger ~~ nil
                ifTrue: [
                  self _commitIndexCreation: indexObj
                    at: i
                    commitInterval: anInteger
                ].
              object := bag _at: i.
              iList addSegment: object segment logging: false.
              " traverse up to the point of the unshared path term "
              object := indexObj traverse: object upTo: pathTerm.
              pathTerm addMappingsForObject: object logging: false
            ]
          ]
          ifFalse: [
            1 to: bag size do: [ :i |
              anInteger ~~ nil
                ifTrue: [
                  self _commitIndexCreation: indexObj
                    at: i
                    commitInterval: anInteger
                ].
              " traverse up to the point of the unshared path term "
              object := indexObj traverse: (bag _at: i) upTo: pathTerm.
              pathTerm addMappingsForObject: object logging: false
            ]
          ]
      ]
  ].
%
category: 'Indexing Support'
method: UnorderedCollection
_asIdentityBag

"Returns the receiver."

^ self
%
category: 'Auditing'
method: UnorderedCollection
_auditAndRepairDepListsLogging: aString

"Get any dependency lists we can find and repair them if they need it.
Report any repairs on the given string."

<primitive: 901>
| depLists sharedDepLists depList cnt |
sharedDepLists := SharedDependencyLists.
depLists := self _getAllDependencyLists.
cnt := 0.
1 to: depLists size do: [ :i |
  (depList := depLists at: i) _needsSorting
    ifTrue: [
      cnt := cnt + 1.
      depList _sortEntries: (sharedDepLists depListBucketFor: depList).
    ]
].
System _disableProtectedMode.
cnt > 0
  ifTrue: [
    aString add: 'For nsc ['; add: self identityHash asString;
      add: '] repaired '; add: cnt asString; add: ' dependency list(s)';
      add: Character lf.
  ].
^ aString
%
category: 'Indexing Audit'
method: UnorderedCollection
_auditIndexes

"Verifies that the index objects are consistent.  This may take a while.
 Returns a string that describes any inconsistencies found.  This method
 is currently unsupported, but is provided for customer support.
 Note that this method may appear much slower than previous releases; this
 is due to a bug that meant the audit skipped all elements." 


<primitive: 901>
| roots sz aString prevObj |
self _indexedPaths == nil
  ifTrue: [
    System _disableProtectedMode.
    ^ 'No indexes are present.'
  ].

roots := self _indexedPaths rootTerms.
sz := roots size.
aString := String new.

" audit by traversing the tree of path terms for each element "
prevObj := #_incompletePathTraversal.
self do: [ :obj | | rootTerm |
  obj ~~ prevObj
    ifTrue: [
      1 to: sz do: [ :j |
        rootTerm := roots at: j.
        rootTerm auditObject: obj
          occurrences: (self occurrencesOf: obj)
          on: aString
      ]
    ].
  prevObj := obj.
].
aString isEmpty
  ifTrue: [ aString := 'Indexes are OK' ].

System _disableProtectedMode.
^ aString
%
category: 'Updating Indexes - Private'
method: UnorderedCollection
_calculateIndexDictionarySize: indexPathLength

"Heuristic for determining size of index dictionary root."

| sz |

" if it has already been set, do not perform heuristic "
self _getIndexDictionaryCreationSize == RcIndexDictionary defaultBasicSize
  ifTrue: [
    sz := Integer _selectedPrimeGreaterThan:
      (((self size * indexPathLength) // 900000) + 1) * 2000.

    self _setIndexDictionaryCreationSize: sz.
  ]
%
category: 'Indexing Support'
method: UnorderedCollection
_checkIndexPathExpression: aPathString

"Raises an error if the given path expression (an Array of Strings), is
 not valid; otherwise returns the receiver."

< protected >

| pathArray pathArraySize tmpClass prevClass pathName |

pathArray := aPathString asArrayOfPathTerms.
pathArraySize := pathArray size.

" currently not allowing set-valued instance variable to be first path term "
(pathArray at: 1) == #* 
  ifTrue: [ ^ self _error: #rtErrFirstPathTermSetValued ].

" single term of empty string is ok "
(pathArraySize == 1 _and: [ (pathArray at: 1) == #'' ])
  ifTrue: [ ^ self ].

1 to: pathArraySize do: [ :i |
  (pathArray at: i) _isValidPathTermName
    ifFalse: [
      ^ self _error: #rtErrInvalidIndexPathExpression
               args: #[ (pathArray at: i) ]
      ]
  ].

tmpClass := self class elementConstraint.
1 to: pathArraySize do: [ :i |
  pathName := pathArray at: i.
  tmpClass == Object
    ifTrue: [ ^ self ].

  prevClass := tmpClass.
  ( tmpClass isNsc _and: [ pathName == #* ] )
    ifTrue: [ tmpClass := tmpClass elementConstraint ]
    ifFalse: [ tmpClass := (tmpClass _constraintOn: pathName) ].

  tmpClass == nil
    ifTrue: [
      ^ self _error: #rtErrPathNotTraversable
               args: #[ prevClass, pathName ]
    ]
  ]

%
category: 'Updating Indexes - Private'
method: UnorderedCollection
_cleanUpDependencies

"The receiver is having all of its indexes removed.  Remove dependency
 list entries for objects along the path of the indexes."

< protected >

| rootTerms bag prevObj obj iList |
iList := self _indexedPaths.
rootTerms := iList rootTerms.

prevObj := #_incompletePathTraversal.
bag := self _asIdentityBag.
1 to: bag size do: [ :i |
  obj := bag _at: i.
  (obj == prevObj)
    ifFalse: [
      1 to: rootTerms size do: [ :j |
        (rootTerms at: j) cleanupDependencyListFor: obj
      ]
    ].
  prevObj := obj
].

iList hasTrackingObjects
  ifTrue: [ iList removeAllIndexesFor: self ]
  ifFalse: [ self _indexedPaths: nil ].

" resize the Array of path terms to avoid object audit errors "
rootTerms size: 0.
%
category: 'Indexing Support'
method: UnorderedCollection
_clearIndexList

"Private.  Sets the indexed paths for the receiver to nil."

self _indexedPaths: nil
%
category: 'Modification Tracking'
method: UnorderedCollection
_clearModificationTrackingTo: tracker

"Removes the given tracker from the receiver's indexedPaths and dependency
 lists."

<primitive: 901>
super _clearModificationTrackingTo: tracker.
self _getIndexList _performNoDebug: #removeTracker:for:
  with: tracker
  with: self.
System _disableProtectedMode.
%
category: 'Clustering'
method: UnorderedCollection
_clusterIndexes

"Clusters indexing objects.  This may cause concurrency conflicts on
 internal indexing objects.  Returns the receiver."

<primitive: 901>
| iList |
(iList := self _indexedPaths) == nil
  ifTrue: [ 
    System _disableProtectedMode.
    ^ self 
  ].

" cluster the index dictionary "
self _indexDictionary ~~ nil
  ifTrue: [ self _indexDictionary clusterDepthFirst ].

" cluster each index object "
1 to: iList size by: 2 do: [ :j |
  (iList at: j) _isIndexObject
    ifTrue: [ (iList at: j) clusterDepthFirst ]
].

System _disableProtectedMode.
%
category: 'Updating Indexes - Private'
method: UnorderedCollection
_commitIndexCreation: indexObj at: offset commitInterval: anInteger

"Commits the current transaction based upon the current offset into the
 receiver."

(offset \\ anInteger) == 0
  ifTrue: [
    indexObj progress: offset.
    System commitTransaction ifFalse: 
        [self _errorCouldNotCommitDuringIndexCreation].
    System transactionMode == #manualBegin ifTrue:
        [System beginTransaction]
  ]
%
category: 'Updating Indexes - Private'
method: UnorderedCollection
_createEqualityIndexOn: aPathString commitInterval: anInteger

"Creates an equality index on aPathString.  Returns nil if the index was
 created.  If there was an error, returns a two-element Array containing the
 error symbol and an Array of arguments."

< protected >

| lastConstraint |
" get the constraint of the last element along the path
( will raise an error if path is not constrained) "
lastConstraint := self getLastElementConstraintOnPath: aPathString.
lastConstraint == nil
  ifTrue: [
    ^ #[ #rtErrBagNoConstraintAlongPath, #[ aPathString ] ]
  ].

^ self _createEqualityIndexOn: aPathString
  withLastElementClass: lastConstraint
  commitInterval: anInteger
%
category: 'Updating Indexes - Private'
method: UnorderedCollection
_createEqualityIndexOn: aPathString
withLastElementClass: aClass
commitInterval: anInteger

"Creates an equality index on aPathString.  Returns nil if the index was
 created.  If there was an error, returns a two-element Array containing the
 error Symbol and an Array of arguments."

< protected >

| indexObj pathArray original updateIndexListSegments |

" check if the last constraint is appropriate Object "
( ( aClass canUnderstand: #<= ) _or:
[ aClass == Boolean _or: [ aClass == UndefinedObject ] ] )
  ifFalse: [
    ^ #[ #rtErrBagClassDoesNotSupportRangeOperators, #[ aClass ] ]
  ].

" see if incomplete index for same path "
( (indexObj := self _findIncompleteIndex) ~~ nil _and:
[ indexObj pathComponentsString = aPathString ] )
  ifTrue: [ ^ nil ].

Exception
  category: nil
  number: nil
  do: [ :ex :cat :num :args | | txt |
    " get the text for the raised error "
    txt := cat textForError: num args: args.
    " check for recursive signal "
    num == (ErrorSymbols at: #rtErrPreventingCommit)
      ifTrue: [ " remove this exception and resignal "
        anInteger ~~ nil
          ifTrue: [ self removeIncompleteIndex ].
        ex remove.
        ex resignal: cat number: num args: args
      ]
      ifFalse: [ " append additional message to the end of text "
        txt _error: #rtErrPreventingCommit
      ]
  ].

pathArray := aPathString asArrayOfPathTerms.
" check if a path string is already used as an index "
(self _findIndexesWithPath: pathArray) do: [ :iObj |
  iObj isRangeEqualityIndex ifTrue: [ ^ nil ]
].

" create the index object "
indexObj := RangeEqualityIndex newWithLastElementClass: aClass.
indexObj moveToDisk.

" don't need index dictionary if path length is one "
pathArray size == 1
  ifFalse: [
    self _calculateIndexDictionarySize: pathArray size - 1.
    indexObj indexDictionary: self _getIndexDictionary
  ].
indexObj nscRoot: self.

updateIndexListSegments :=
  self _indexedPaths == nil _or: [ self _indexedPaths _hasExplicitIndex not ].

self _getIndexList buildPathTermsFor: indexObj with: pathArray.
self _putInWriteSet.
indexObj changeToSegment: GsIndexingSegment.

original := indexObj preIndexCreation.

" optimize creation of indexes on elements of the NSC itself "
indexObj isIndexOnRootNsc
  ifTrue: [
    indexObj addDirectMappingsFor: self _asIdentityBag
      indexList: self _indexedPaths
      updateIndexListSegments: updateIndexListSegments
      commitInterval: anInteger
  ]
  ifFalse: [
    self _addMappingsFor: indexObj
      updateIndexListSegments: updateIndexListSegments
      commitInterval: anInteger
  ].

indexObj postIndexCreation: original commitInterval: anInteger.

^ nil
%
category: 'Updating Indexes - Private'
method: UnorderedCollection
_createIdentityIndexOn: aPathString commitInterval: anInteger

"Creates an identity index on aPathString.  Returns the receiver."

< protected >

| indexObj indexes pathArray lastConstraint original updateIndexListSegments |

" see if incomplete index for same path "
( (indexObj := self _findIncompleteIndex) ~~ nil _and:
[ indexObj pathComponentsString = aPathString ] )
  ifTrue: [ ^ self ].

pathArray := aPathString asArrayOfPathTerms.
" check if an identity index already exists for the path "
indexes := self _findIndexesWithPath: pathArray.
1 to: indexes size do: [ :i |
  (indexes at: i) isIdentityIndex ifTrue: [ ^ self ]
].

lastConstraint := self getLastElementConstraintOnPath: aPathString.

( lastConstraint ~~ nil _and: [ self _isRangeable: lastConstraint ] )
  ifTrue: [
    1 to: indexes size do: [ :i |
      (indexes at: i) isRangeEqualityIndex ifTrue: [ ^ nil ]
    ].
    indexObj := RangeEqualityIndex newWithLastElementClass: lastConstraint
  ]
  ifFalse: [ indexObj := IdentityIndex new ].
indexObj moveToDisk.

" don't need index dictionary if path length is one "
( pathArray size == 1 _and: [ indexObj isRangeEqualityIndex ] )
  ifFalse: [
    self _calculateIndexDictionarySize: pathArray size.
    indexObj indexDictionary: self _getIndexDictionary
  ].
indexObj nscRoot: self.

updateIndexListSegments := self _indexedPaths == nil.
self _getIndexList buildPathTermsFor: indexObj with: pathArray.
self _putInWriteSet.
indexObj changeToSegment: GsIndexingSegment.

original := indexObj preIndexCreation.

" optimize creation of indexes on elements of the NSC itself "
indexObj isIndexOnRootNsc
  ifTrue: [
    indexObj addDirectMappingsFor: self _asIdentityBag
      indexList: self _indexedPaths
      updateIndexListSegments: updateIndexListSegments
      commitInterval: anInteger
  ]
  ifFalse: [
    self _addMappingsFor: indexObj
      updateIndexListSegments: updateIndexListSegments
      commitInterval: anInteger
  ].

indexObj postIndexCreation: original commitInterval: anInteger.
%
category: 'Updating Indexes - Private'
method: UnorderedCollection
_deadNscBeingGarbageCollected

"The receiver is a not connected object being garbage collected.  This
 method implements an optimization that does not update dependency lists,
 but instead lets garbage collection and lazy updating keep objects
 consistent.  This also prevents the GcGem from causing concurrency
 conflicts on the elements that participate in the index."

| iList indexObj |

iList := self _indexedPaths.
" scan indexed paths for explicitly created indexes "
1 to: iList size by: 2 do: [ :i |
  (iList at: i + 1) == 1
    ifTrue: [
      indexObj := iList at: i.
      1 to: indexObj size do: [ :j |
        " just clear the path term "
        (indexObj at: j) _clear.
      ]
    ]
].
%
category: 'Accessing Indexes'
method: UnorderedCollection
_explicitEqualityIndexedPaths

"Returns an Array of Strings, each of which represents a path for which an
 explicitly created equality index exists in the receiver.  Each path
 originates with the elements of the receiver."

<primitive: 901>
| anArray |
anArray := Array new.
self _indexedPaths == nil
  ifTrue: [
    System _disableProtectedMode.
    ^ anArray
  ].
self _indexedPaths indexObjectsAndOffsetsDo: [ :indexObj :offset |
  ( indexObj isRangeEqualityIndex _and: [ offset == 1 ] )
    ifTrue: [
      anArray addLast: (indexObj pathComponentsStringStartingAt: offset)
    ]
].
System _disableProtectedMode.
^ anArray
%
category: 'Accessing Indexes'
method: UnorderedCollection
_explicitEqualityIndexedPathsAndConstraints

"Returns an Array containing info about explicitly created equality indexes.
 The Array consists of String/Class pairs.  The string represents a path of
 the receiver's element kind for which an equality index exists in the
 receiver.  The class is the constraint on the last element in the path."

<primitive: 901>
| anArray |
anArray := Array new.
self _indexedPaths == nil
  ifTrue: [
    System _disableProtectedMode.
    ^ anArray
  ].
self _indexedPaths indexObjectsAndOffsetsDo: [ :indexObj :offset |
  ( indexObj isRangeEqualityIndex _and: [ offset == 1 ] )
    ifTrue: [
      anArray addLast: (indexObj pathComponentsStringStartingAt: offset).
      anArray addLast: indexObj lastElementClass
    ]
].
System _disableProtectedMode.
^ anArray
%
category: 'Accessing Indexes'
method: UnorderedCollection
_explicitIdentityIndexedPaths

"Returns an Array of Strings, each of which represents a path for which an
 explicitly created identity index exists in the receiver.  Each path
 originates with the elements of the receiver."

<primitive: 901>
| anArray |
anArray := Array new.
self _indexedPaths == nil
  ifTrue: [
    System _disableProtectedMode.
    ^ anArray
  ].

self _indexedPaths indexObjectsAndOffsetsDo: [ :indexObj :offset |
  offset == 1
    ifTrue: [
      indexObj isIdentityIndex
        ifTrue: [ | pathString |
          pathString := indexObj pathComponentsStringStartingAt: offset.
          (anArray includesValue: pathString)
            ifFalse: [ anArray addLast: pathString ]
        ]
        ifFalse: [
          (self _isRangeable: indexObj lastElementClass)
            ifTrue: [ | pathString |
              pathString := indexObj pathComponentsStringStartingAt: offset.
              (anArray includesValue: pathString)
                ifFalse: [ anArray addLast: pathString ]
            ]
        ]
    ]
].
System _disableProtectedMode.
^ anArray
%
category: 'Updating Indexes - Private'
method: UnorderedCollection
_findIncompleteIndex

"Returns an index that is incomplete, or nil if not found."

| iList |
iList := self _indexedPaths.
iList == nil
  ifTrue: [ ^ nil ].

" for each index on the NSC ... "
1 to: iList size by: 2 do: [ :i |
  " only look at the index if the receiver is the root "
  (iList at: i + 1) == 1
    ifTrue: [
      " see if the index is complete "
      (iList at: i) isComplete
        ifFalse: [ ^ iList at: i ]
    ]
].
^ nil
%
category: 'Indexing Support'
method: UnorderedCollection
_findIndexesWithPath: pathArray

"Returns an Array of index objects whose path components are the same as
 represented by the path string.  This method only considers those indexes that
 are defined with the receiver as the root NSC."

< protected >

| anArray iList |
iList := self _indexedPaths.
iList == nil
  ifTrue: [ ^ Array new ].

anArray := Array new.
" for each index on the NSC ... "
1 to: iList size by: 2 do: [ :i |
  " only look at the index if the receiver is the root "
  (iList at: i + 1) == 1
    ifTrue: [
      " if the index is on the same path, add it to the Array "
      ((iList at: i) hasIndexOnPath: pathArray)
        ifTrue: [ anArray addLast: (iList at: i) ]
    ]
].
^ anArray
%
category: 'Indexing Support'
method: UnorderedCollection
_findRangeIndexWithPath: pathArray

"Returns a range equality index whose path components are the same as
 represented by the pathArray (Array of strings)."

< protected >

| index iList |
iList := self _indexedPaths.
iList == nil
  ifTrue: [ ^ nil ].

" for each index on the NSC ... "
1 to: iList size by: 2 do: [ :i |
  index := iList at: i.
  ( index _isIndexObject _and:
  [ index isRangeEqualityIndex _and:
  [ index hasIndexOnPath: pathArray ] ] )
    ifTrue: [ ^ index ]
].
^ nil
%
category: 'Private'
method: UnorderedCollection
_finishShallowCopy

"Private."

_indexedPaths := nil
%
category: 'Auditing'
method: UnorderedCollection
_getAllDependencyLists

"Return a set of all dependency lists that we can find from the receiver."

| all iList rootTerms classSystem obj sz prevObj bag |

all := IdentitySet new.
iList := self _indexedPaths.
bag := self _asIdentityBag.
iList ~~ nil
  ifTrue: [
    classSystem := System.
    prevObj := #_incompletePathTraversal.
    rootTerms := iList rootTerms.
    sz := rootTerms size.
    1 to: bag size do: [ :i |
      obj := bag _at: i.
      obj == prevObj
        ifFalse: [
          1 to: sz do: [ :j |
            (rootTerms at: j) allDependencyListsFor: obj into: all
          ]
        ].
      prevObj := obj
    ]
  ].
^ all
%
category: 'Indexing Support'
method: UnorderedCollection
_getConstraintsOnPath: pathArray

"Returns an Array of Classes consisting of the constraint on unnamed instance
 variables of the receiver, followed by the constraints for each instance
 variable in pathArray (an Array of Strings).  If any of the Classes that are
 traversed is not constrained on the given instance variable name, the
 remaining constraints in the Array are nil."

< protected >

| tmpClass constraintArray pathName sz |
sz := pathArray size.
constraintArray := Array new: sz.

" if path is empty string, then just use the constraint on unnamed variables "
(sz == 1 _and: [ (pathArray at: 1) == #'' ] )
  ifTrue: [
    constraintArray at: 1 put: self class elementConstraint.
    ^ constraintArray
  ].

" get constraint on unnamed variables for root NSC "
tmpClass := self class elementConstraint.
1 to: sz do: [ :i |
  constraintArray at: i put: tmpClass.
  pathName := pathArray at: i.
  tmpClass == nil
    ifFalse: [
      tmpClass isNsc
        ifTrue: [ tmpClass := tmpClass elementConstraint ]
        ifFalse: [ tmpClass := (tmpClass _constraintOn: pathName) ]
    ].
].
^ constraintArray
%
category: 'Indexing Support'
method: UnorderedCollection
_getIndexDictionary

"Returns the index dictionary that is shared by all indexes.  If there are no
 indexes, create an index dictionary."

| iList iDict |
iList := self _indexedPaths.
iList == nil
  ifTrue: [
    iDict := RcIndexDictionary new: self _getIndexDictionaryCreationSize.
  ]
  ifFalse: [
    1 to: iList size by: 2 do: [ :i |
      " only use the index dictionary if the receiver is the root
        NSC of the index "
      ( (iList at: i + 1) == 1 _and:
      [ (iList at: i) indexDictionary ~~ nil ] )
        ifTrue: [ ^ (iList at: i) indexDictionary ]
    ].
    iDict := RcIndexDictionary new: self _getIndexDictionaryCreationSize.
  ].
iDict changeToSegment: GsIndexingSegment.
iDict moveToDisk.
^ iDict
%
category: 'Indexing Support'
method: UnorderedCollection
_getIndexDictionaryCreationSize

"Returns the basic size of an index dictionary when a new one is
 being created."

< protected >


^ System
  rcValueCacheAt: #indexDictionarySize
  for: self
  otherwise: RcIndexDictionary defaultBasicSize.
%
category: 'Indexing Support'
method: UnorderedCollection
_getIndexList

"Returns the index list for the receiver.  If one does not exist, create it."

< protected >

| iList |
iList := self _indexedPaths.
iList == nil
  ifTrue: [
    iList := IndexList new.
    self _indexedPaths: iList.

    " don't put in bitmap if it is a permanent object "
    self isCommitted
      " set WorkNscsWithNewIdxs bitmap for this NSC "
      ifFalse: [ System _add: self to: 23 ].

    iList moveToDisk.
    " move receiver to disk so we don't have tmpNscsWithIdxs "
    self moveToDisk.

    iList changeToSegment: GsIndexingSegment.
  ].

^ iList
%
category: 'Indexing Support'
method: UnorderedCollection
_getIndexReferencesInto: refsToRcvr

"Returns an Array of information about references to the receiver due to the
 receiver's participation in an index as a set-valued instance variable.
 The Array consists of pairs:

 1) An object that references the receiver.
 2) The offset of the instance variable in that object that references the
  receiver.

 This is used only in support of the 'become:' method; therefore if
 the receiver is a root NSC with an index, an error will be raised."

< protected >

| indexList pathTerm index vals obj pathTerms |
indexList := self _indexedPaths.
pathTerms := IdentitySet new.
" for each index that the receiver participates in ... "
2 to: indexList size by: 2 do: [ :i |
  " check if it is the root NSC or a set-valued instance variable "
  (indexList at: i) == 1
    ifTrue: [ " it is the root, 'become:' not allowed "
      self _error: #rtErrCantBecomeIndexedNsc.
    ]
    ifFalse: [
      (indexList at: i) > 1
        ifTrue: [
          index := indexList at: i - 1.
          pathTerm := index at: (indexList at: i) - 2.
          " only want to check each unique path term "
          (pathTerms includesIdentical: pathTerm)
            ifFalse: [
              " get the objects that refer to the receiver "
              vals := index indexDictionary at: self term: pathTerm otherwise: nil.
              nil == vals
                ifTrue: [
                  index indexDictionary _errorNoEntryForKey: self
                    term: pathTerm
                    value: nil
                ].

              (BucketValueBag _hasInstance: vals)
                ifTrue: [
                  1 to: vals size do: [ :j |
                    obj := vals _at: j.
                    pathTerm indicatesNsc
                      ifTrue: [
                        self _addToReferences: obj
                          offset: 0
                          occurrences: (obj occurrencesOf: self)
                          into: refsToRcvr
                      ]
                      ifFalse: [
                        self _addToReferences: obj
                          offset: (pathTerm _ivOffsetFor: obj)
                          occurrences: 1
                          into: refsToRcvr
                      ].
                  ]
                ]
                ifFalse: [
                  pathTerm indicatesNsc
                    ifTrue: [
                      self _addToReferences: vals
                      offset: 0
                      occurrences: (vals occurrencesOf: self)
                      into: refsToRcvr
                    ]
                    ifFalse: [
                      self _addToReferences: vals
                      offset: (pathTerm _ivOffsetFor: vals)
                      occurrences: 1
                      into: refsToRcvr
                    ]
                ].
                pathTerms add: pathTerm
              ]
        ]
    ]
].
" since NSCs may have named instance variables, invoke this on super "
^ super _getIndexReferencesInto: refsToRcvr
%
category: 'Testing'
method: UnorderedCollection
_hasImplicitIndexOn: pathArray

"Returns whether the receiver has any implicit index on the given path."

< protected >

| iList |
iList := self _indexedPaths.
" for each index ... "
1 to: iList size by: 2 do: [ :i |
  " if it is not an explicit index and the remaining path matches "
  ( (iList at: i + 1) > 1 _and:
  [ (iList at: i) hasIndexOnPath: pathArray startingAt: (iList at: i + 1)] )
    ifTrue: [ ^ true ]
].
^ false
%
category: 'Indexing Support'
method: UnorderedCollection
_hasIndexes

"Returns whether the receiver has any indexes on it."

<primitive: 901>
| result iList |
iList := self _indexedPaths.
result := false.
iList == nil
  ifFalse: [
    2 to: iList size by: 2 do: [ :i |
      (iList at: i) > 0 ifTrue: [ result := true ]
    ]
  ].
System _disableProtectedMode.
^ result
%
category: 'Searching'
method: UnorderedCollection
_idxOccurrencesOf: aValue

"Returns the number of occurrences of the given value in the receiver."

^ self occurrencesOf: aValue
%
category: 'Indexing Support'
method: UnorderedCollection
_indexDictionary

"Returns the index dictionary that is shared by all indexes.  If there are no
 indexes, returns nil."

< protected >

| iList |
iList := self _indexedPaths.
iList == nil
  ifTrue: [ ^ nil ]
  ifFalse: [
    1 to: iList size by: 2 do: [ :i |
      ( (iList at: i + 1) == 1 _and:
      [ (iList at: i) indexDictionary ~~ nil ] )
        ifTrue: [ ^ (iList at: i) indexDictionary ]
    ].
    ^ nil
  ]
%
category: 'Indexing Support'
method: UnorderedCollection
_indexedPaths

"Private.  For use only by the indexing subsystem.

 Returns the indexed paths for the receiver."

^ _indexedPaths
%
category: 'Indexing Support'
method: UnorderedCollection
_indexedPaths: anIndexList

"Private.  For use only by the indexing subsystem.

 Sets the indexed paths for the receiver."

<primitive: 361>
self _primitiveFailed: #_indexedPaths: .
self _uncontinuableError
%
category: 'Indexing Support'
method: UnorderedCollection
_indexParticipationInfoInto: array

"Returns an Array of pairs:  the root NSC and the path string that describes
 the path traversed from the root to reach the receiver."

< protected >

| iList indexObj pathString included |
iList := self _indexedPaths.
1 to: iList size by: 2 do: [ :i |
  (iList at: i + 1) > 1
    ifTrue: [
      indexObj := iList at: i.
      pathString := indexObj _partialPathComponentsStringUpTo: (iList at: i + 1).
      included := false.
      1 to: array size by: 2 do: [ :j |
        ( indexObj nscRoot == (array at: j) _and:
        [ (array at: j + 1) = pathString ] )
          ifTrue: [ included := true ]
      ].
      included
        ifFalse: [
          array add: indexObj nscRoot.
          array add: pathString
        ]
    ]
].
^ super _indexParticipationInfoInto: array
%
category: 'Indexing Support'
method: UnorderedCollection
_indexStatistics

"Return a dictionary containing statistics that can be useful in determining the
 performance of indexes.  This information is useful for system administrators
 and customer support in determining how to optimize index usage."

< primitive: 901 >
| dict |
dict := SymbolDictionary new.
self _indexedPaths == nil
  ifTrue: [
    dict at: #NoIndexesPresent put: true.
    System _disableProtectedMode.
    ^ dict
  ].

self _indexedPaths _statisticsInto: dict.
self _indexDictionary ~~ nil
  ifTrue: [ self _indexDictionary _statisticsInto: dict ].

System _disableProtectedMode.
^ dict
%
category: 'Testing'
method: UnorderedCollection
_isIdentityBag

"Return whether the receiver is an identity-based collection."

^ false
%
category: 'Indexing Support'
method: UnorderedCollection
_isLastOccurrenceInIndexObjects: anObject

"Returns true if the given object is maintained in the indexing objects for one
 occurrence."

< protected >

| rootTerms pathTerm key val tmpList val2 num parentTerm |
rootTerms := self _indexedPaths rootTerms.
" find a path term with a mapping in the index dictionary "
1 to: rootTerms size do: [ :i |
  pathTerm := rootTerms at: i.
  pathTerm _isObsoletePathTerm
    ifFalse: [
      pathTerm isRangeEqualityIndexLastPathTerm
        ifTrue: [
          pathTerm offset == 1
            ifTrue: [ 
              " if index directly on NSC elements or anObject is nil "
              ( pathTerm indicatesIndexOnNscElements _or: [ nil == anObject ] )
                ifTrue: [ key := anObject ]
                ifFalse: [ key := pathTerm _nextObjectFor: anObject ].
              tmpList := IdentityBag new.

              pathTerm updateBtree btreeRoot _findAllValuesFor: anObject identicalKey: key into: tmpList.
              ^ (tmpList occurrencesOf: anObject) <= 1
            ]
            ifFalse: [
              pathTerm := pathTerm getParentTerm.
              val := pathTerm updateDict at: anObject term: pathTerm otherwise: nil.
              " see if more than one mapping "
              (BucketValueBag _hasInstance: val)
                ifTrue: [
                  pathTerm indicatesNsc
                    ifTrue: [
                      num := val occurrencesOf: self.
                      " if anObject is contained in other NSCs "
                      num < val size
                        ifTrue: [ ^ false ].

                      " get path term before this one "
                      pathTerm := pathTerm getParentTerm.
                      val2 := pathTerm updateDict at: self term: pathTerm otherwise: nil.
                      (BucketValueBag _hasInstance: val2)
                        ifTrue: [ ^ val2 size == num ]
                        ifFalse: [ ^ true ]
                    ]
                    ifFalse: [
                      ^ (val occurrencesOf: self) <= 1
                    ]
                ]
                ifFalse: [ ^ true ]
            ]
        ]
        ifFalse: [
          " get key to look up in index dictionary "
          ( pathTerm indicatesIndexOnNscElements _or: [ nil == anObject ] )
            ifTrue: [ key := anObject ]
            ifFalse: [
              " see if a path with '*.*' in it "
              pathTerm indicatesNsc
                ifTrue: [
                  val := pathTerm updateDict at: anObject
                    term: pathTerm getParentTerm
                    otherwise: nil.
                  ^ (BucketValueBag _hasInstance: val) not
                ]
                ifFalse: [ key := pathTerm _nextObjectFor: anObject ]
            ].
          " look up the mapping in the index dictionary "
          val := pathTerm updateDict at: key term: pathTerm otherwise: nil.
          " see if more than one mapping "
          (BucketValueBag _hasInstance: val)
            ifTrue: [
              (num := val occurrencesOf: anObject) <= 1
                ifTrue: [ ^ true ]
                ifFalse: [
                  " see if multiple occurrences are due to more than one object
                    referencing self "
                  " if there is a parent term, it is a SetValuedPathTerm "
                  parentTerm := pathTerm getParentTerm.
                  parentTerm == nil
                    ifTrue: [ ^ false ].

                val2 := pathTerm updateDict at: anObject term: parentTerm otherwise: nil.
                (BucketValueBag _hasInstance: val2)
                  ifTrue: [ ^ (val2 occurrencesOf: self) == num ]
                  ifFalse: [ ^ false ]
                ]
            ]
            ifFalse: [ ^ true ]
        ]
    ]
].
" if we get this far, there are only incomplete indexes "
^ true
%
category: 'Indexing Support'
method: UnorderedCollection
_isLastOccurrenceInIndexObjectsFor: anObject
  "Returns true if the given object is maintained in the indexing objects for one
   occurrence. For identitySet and Set this method can be overridden to return true."
^self _isLastOccurrenceInIndexObjects: anObject
%
category: 'Indexing Support'
method: UnorderedCollection
_isRangeable: aClass

"Returns whether the given class should have an equality index created when an
 identity index is requested."

< protected >

^ aClass _isSpecial _and: [ (aClass isSubclassOf: AbstractCharacter) not ]
%
category: 'Updating Indexes - Private'
method: UnorderedCollection
_lockForIndexCreation

"Locks the receiver to prevent concurrent users from invalidating index
 creation."

| lockList |
lockList := Array with: self.
self _indexDictionary ~~ nil
  ifTrue: [
    self _indexDictionary _doCollisionBuckets: [ :cb | lockList add: cb ]
  ].
System writeLockAll: lockList.
System addAllToCommitOrAbortReleaseLocksSet: lockList.
%
category: 'Updating Indexes - Private'
method: UnorderedCollection
_primRecomputeIndexSegments

""

<protected primitive: 214>
self _primitiveFailed: #_primRecomputeIndexSegments .
self _uncontinuableError
%
category: 'Indexing Support'
method: UnorderedCollection
_putInWriteSet

"In some cases, it is necessary to put an NSC with indexes into the write set
 explicitly.  For example, when NO_RW_CHECKS concurrency mode is enabled, index
 creation should put the NSC in the write set to ensure that other transactions
 do not commit additions to the NSC successfully (and thus corrupting the
 internal indexing objects).  This method puts the receiver in the write set by
 writing the _indexedPaths."

self _indexedPaths: self _indexedPaths
%
category: 'Indexing Audit'
method: UnorderedCollection
_quickCheckIndexes

"Verifies that the index objects are present. This method is does 
 not catch most types of index corruption. The return values are 
 messages that no indexes are present, or that indexes are okay. 
 Any errors found will result in an unhandled exception. 
 
 This method is unsupported, but is provided for customer support."

<primitive: 901>
| prevObj |

self _indexedPaths == nil
  ifTrue: [
    System _disableProtectedMode.
    ^ 'No indexes are present.'
  ].

"execute minor code over all the path terms"
self _indexedPaths rootTerms do: 
  [:rootTerm |
  (rootTerm _thisAndAllChildTerms) do:
     [:aTerm | 
     aTerm indicatesNsc
       ifFalse: [aTerm isRangeEqualityIndexLastPathTerm]]].

prevObj := #_incompletePathTraversal.
self do: 
  [ :obj | 
  obj ~~ prevObj
     ifTrue: [ obj _indexParticipationInfo].
  prevObj := obj].

System _disableProtectedMode.

^'Indexes are OK'
%
category: 'Error Handling'
method: UnorderedCollection
_raiseIndexingError: errArray

"Raises an error that was due to updating indexes.  Information
 about the error is contained in the given error Array."

| errNum indexesModified |

indexesModified := errArray at: 1.
errNum := errArray at: 2.
errArray removeFrom: 1 to: 2.
indexesModified
  ifTrue: [
    ^ System
        signal: (ErrorSymbols at: #rtErrPreventingCommit)
        args: errArray
        signalDictionary: GemStoneError
  ]
  ifFalse: [
    ^ System
        signal: errNum
        args: errArray
        signalDictionary: GemStoneError
  ]
%
category: 'Updating Indexes - Private'
method: UnorderedCollection
_recomputeIndexSegments

""

< protected >

self _indexedPaths _emptySegmentSets.
self _primRecomputeIndexSegments.
%
category: 'Indexing Support'
method: UnorderedCollection
_refreshIndexCaches

"This method is currently unsupported, but is provided for customer support.

 Verifies that the index dictionary bucket caches are correct.  If an invalid
 entry is found, fixes it and reports where the fix occurred."

< primitive: 901 >
| result |
self _indexDictionary ~~ nil
  ifTrue: [ result := self _indexDictionary _refreshBucketCaches ].
System _disableProtectedMode.
( result == nil _or: [ result isEmpty ] )
  ifTrue: [ result := 'No problems found with index caches' ].
^ result
%
category: 'Updating Indexes - Private'
method: UnorderedCollection
_removeAllIncompleteIndexes

" Remove index list entries for any incomplete indexes found in the receiver.
Returns the number of incomplete indexes found."

<primitive: 901>
| iList incompletes |

iList := self _indexedPaths.
iList == nil
  ifTrue: [
    System _disableProtectedMode.
    ^ 0
  ].

" gather all incomplete indexes "
1 to: iList size by: 2 do: [ :i |
  ((iList at: i + 1) > 0 _and: [ (iList at: i) isComplete ])
    ifFalse: [
      " lazy instantiation (to ease GC) "
      incompletes == nil
        ifTrue: [ incompletes := Array new ].
      incompletes add: (iList at: i); add: (iList at: i + 1)
    ]
].

1 to: incompletes size by: 2 do: [ :i |
  iList removeIndex: (incompletes at: i)
    withOffset: (incompletes at: i + 1)
    for: self.
].

System _disableProtectedMode.
^ incompletes size / 2
%
category: 'Updating Indexes - Private'
method: UnorderedCollection
_removeAllRootIndexes: rootIndexes hasNonRootIndex: hasNonRootIndex

"Removes all root indexes on the receiver."

< protected >

| array roots |

hasNonRootIndex
  ifTrue: [
    " only remove root indexes "
    1 to: rootIndexes size do: [ :i |
      self _removeIndex: (rootIndexes at: i)
    ].
    " if the receiver still has implicit indexes ... "
    self _indexedPaths == nil
      ifFalse: [ ^ self _indexParticipationInfo ]
  ]
  ifFalse: [
    array := Array new.
    roots := self _indexedPaths rootTerms.
    1 to: roots size do: [ :i |
      (roots at: i) _thisAndAllChildTermsInto: array
    ].

    " this will clean up dependency tags and nil out the index list "
    self _cleanUpDependencies.

    " remove dependency lists from global table that are no longer needed "
    SharedDependencyLists removeEntriesContaining: array.
  ].
%
category: 'Updating Indexes - Private'
method: UnorderedCollection
_removeEqualityIndexOn: aPathString

"Removes an equality index on aPathString.  Returns aPathString if the
index is implicit; returns nil if there is no index along the given path."

< protected >

| indexObj indexes pathArray iList |
iList := self _indexedPaths.
pathArray := aPathString asArrayOfPathTerms.
" check if an index exists for the path string "
indexes := self _findIndexesWithPath: pathArray.
indexes isEmpty
  ifTrue: [ " see if there was an implicit index for aPathString "
    (self _hasImplicitIndexOn: pathArray)
      ifTrue: [ ^ aPathString ]
  ]
  ifFalse: [
    " make sure it is an equality index "
    1 to: indexes size do: [ :i |
      (indexes at: i) isRangeEqualityIndex
        ifTrue: [ indexObj := (indexes at: i) ]
    ].
    " if it wasn't, then see if it has an implicit index "
    indexObj == nil
      ifTrue: [
        (self _hasImplicitIndexOn: pathArray)
          ifTrue: [ ^ aPathString ]
          ifFalse: [ ^ nil ]
      ].
  ].

" no index was found with the given path "
indexObj == nil
  ifTrue: [ ^ nil ].

Exception
  category: nil
  number: nil
  do: [ :ex :cat :num :args | | txt |
  " get the text for the raised error "
  txt := cat textForError: num args: args.
  " check for recursive signal "
  num == (ErrorSymbols at: #rtErrPreventingCommit)
    ifTrue: [ " remove this exception and resignal "
      ex remove.
      ex resignal: cat number: num args: args
    ]
    ifFalse: [ " append additional message to the end of text "
      txt _error: #rtErrPreventingCommit
    ]
  ].

self _removeIndex: indexObj.
%
category: 'Updating Indexes - Private'
method: UnorderedCollection
_removeIdentityIndexOn: aPathString

"Removes an identity index on aPathString.  Returns aPathString if the
index is implicit; returns nil if there is no index along the given path."

< protected >

| indexObj indexes pathArray iList |
iList := self _indexedPaths.
pathArray := aPathString asArrayOfPathTerms.
" check if an index exists for the path string "
indexes := self _findIndexesWithPath: pathArray.

indexes isEmpty
  ifTrue: [ " see if there was an implicit index for aPathString "
    (self _hasImplicitIndexOn: pathArray)
      ifTrue: [ ^ aPathString ]
  ]
  ifFalse: [
    " make sure it is an identity index "
    indexes do: [ :iObj |
      ( iObj isIdentityIndex _or:
      [ self _isRangeable: iObj lastElementClass ] )
        ifTrue: [ indexObj := iObj ]
    ].
    " if it wasn't, then see if it has an implicit index "
    indexObj == nil
      ifTrue: [
        (self _hasImplicitIndexOn: pathArray)
          ifTrue: [ ^ aPathString ]
          ifFalse: [ ^ nil ]
      ]
  ].
" no index was found with the given path "
indexObj == nil
  ifTrue: [ ^ nil ].

Exception
  category: nil
  number: nil
  do: [ :ex :cat :num :args | | txt |
  " get the text for the raised error "
  txt := cat textForError: num args: args.
  " check for recursive signal "
  num == (ErrorSymbols at: #rtErrPreventingCommit)
    ifTrue: [ " remove this exception and resignal "
      ex remove.
      ex resignal: cat number: num args: args
    ]
    ifFalse: [ " append additional message to the end of text "
      txt _error: #rtErrPreventingCommit
    ]
  ].

self _removeIndex: indexObj
%
category: 'Updating Indexes - Private'
method: UnorderedCollection
_removeIndex: indexObj

"Removes the index from the receiver.  Returns the receiver."

< protected >

| pathTerm endObj parentTerm obj iList bag objNsc incomplete updateIndexList
allTerms setOffset |

" since indexes may share path terms, find the first one that is not shared "
pathTerm := indexObj _findFirstUnsharedPathTerm.
pathTerm == nil
  ifTrue: [ self _removeSubsumedIndex: indexObj ]
  ifFalse: [
    parentTerm := pathTerm getParentTerm.
    bag := self _asIdentityBag.
    incomplete := #_incompletePathTraversal.

    "See if there is a set-valued instance variable before the unshared path
     term."
    ( (setOffset := indexObj _findSetValuedPathTermOffset) ~~ 0 _and:
    [ setOffset < pathTerm offset ] )
      ifTrue: [
        " collect objects in a Bag to update the index dictionary "
        updateIndexList := false.
        " see if index is subsumed by an existing index "
        ( pathTerm == indexObj lastPathTerm _and:
        [ pathTerm size > 1 _and:
        [ (pathTerm at: 1) size ~= (pathTerm at: 2) size ] ] )
          ifTrue: [
            " if last path term is #*, will need to update the
              last objects' index list "
            pathTerm indicatesNsc
              ifTrue: [
                updateIndexList := true.
                objNsc := NscBuilder for: IdentityBag new max: indexObj sizeForNscBuilder
              ]
              ifFalse: [ objNsc := nil ]
          ]
          ifFalse: [ objNsc := NscBuilder for: IdentityBag new max: indexObj sizeForNscBuilder ].

        " for each element in the NSC ... "
        1 to: bag size do: [ :i |
          " traverse up to the point of the unshared path term "
          indexObj 
            traverse: (bag _at: i) 
            upTo: pathTerm 
            startingAt: 1 
            into: objNsc
            addingIndex: false
        ].
        objNsc ~~ nil
          ifTrue: [
            updateIndexList
              ifTrue: [
                pathTerm removeIndexListEntriesFor: objNsc completeBag index: indexObj
              ]
              ifFalse: [
                objNsc := objNsc completeBag.
                1 to: objNsc size do: [ :j |
                  endObj := objNsc _at: j.
                  " only remove mappings once for a given object "
                  endObj ~~ incomplete
                  ifTrue: [
                    pathTerm removeMappingsFor: endObj
                    lastOne: true
                    logging: false.
                    DependencyList removePathTerm: pathTerm for: endObj
                  ].
                  "incomplete := endObj (comment out to fix bug 9536)"
                ]
              ]
          ]
      ]
      ifFalse: [
        " see if the index is on elements of the NSC "
        pathTerm indicatesIndexOnRootNsc
          ifTrue: [
            1 to: bag size do: [ :j |
              obj := bag _at: j.
              pathTerm removeDirectMappingFor: obj logging: false.
              DependencyList removePathTerm: pathTerm for: obj.
            ]
          ]
          ifFalse: [
            1 to: bag size do: [ :j |
              endObj := indexObj traverse: (bag _at: j) upTo: pathTerm.
              ( nil ~~ endObj _and: [ incomplete ~~ endObj ] )
                ifTrue: [
                  pathTerm removeMappingsFor: endObj
                    lastOne: true
                    logging: false.
                  DependencyList removePathTerm: pathTerm for: endObj
                ]
            ]
          ]
      ].
    allTerms := pathTerm _thisAndAllChildTerms.
  ].

" remove index from each path term "
1 to: indexObj size do: [ :i |
  (indexObj at: i) removeIndex: indexObj
].

" remove dependency lists from global table that are no longer needed "
allTerms ~~ nil
  ifTrue: [
    SharedDependencyLists removeEntriesContaining: allTerms.
  ].

iList := self _indexedPaths.
" remove the entry in the receiver's index list "
iList removeIndex: indexObj withOffset: 1 for: self.

parentTerm ~~ nil
  ifTrue: [ parentTerm _determineChildren ]
%
category: 'Updating Indexes - Private'
method: UnorderedCollection
_removeSubsumedIndex: indexObj

"The given index is subsumed by an existing index and is being removed."

< protected >

| pathTerm bag objNsc |
" see if there is a set-valued term "
indexObj _findSetValuedPathTermOffset == 0
  ifFalse: [ " need to update index lists "
    bag := self _asIdentityBag.
    pathTerm := indexObj lastPathTerm.
    " if last path term is *, need to update the last objects index list "
    pathTerm indicatesNsc
      ifTrue: [ objNsc := NscBuilder for: IdentityBag new max: indexObj sizeForNscBuilder ]
      ifFalse: [ objNsc := nil ].

    " for each element in the NSC ... "
    1 to: bag size do: [ :i |
      indexObj
        traverse: (bag _at: i)
        upTo: pathTerm
        startingAt: 1
        into: objNsc
        addingIndex: false
    ].

    objNsc ~~ nil
      ifTrue: [
        pathTerm removeIndexListEntriesFor: objNsc completeBag index: indexObj
      ]
  ].
%
category: 'Indexing Support'
method: UnorderedCollection
_setIndexDictionaryCreationSize: aNumber

"Sets the basic size of an index dictionary when a new one is being created.
 This will only be in effect for the life of the transaction."

aNumber _validateClass: Number.
System
  rcValueCacheAt: #indexDictionarySize
  put: aNumber
  for: self.
aNumber
%
category: 'Indexing Support'
method: UnorderedCollection
_setIndexDictionaryWorkingSetInterval: aNumber

"Sets the working set interval for the index dictionary when index creation
 is in progress.  This will only be in effect for the life of the transaction."

aNumber _validateClass: Number.
System
  rcValueCacheAt: #indexDictionaryWorkingSetInterval
  put: aNumber
  for: self.
^ aNumber
%
category: 'Modification Tracking'
method: UnorderedCollection
_setModificationTrackingTo: tracker

"Adds the given tracker to the receiver's indexedPaths and dependency lists."

<primitive: 901>
| removeFromHiddenSet |
removeFromHiddenSet := false.
" determine if need to remove from WorkNscsWithNewIdxs bitmap "
(self isCommitted not _and: [ self _indexedPaths == nil ])
  ifTrue: [ removeFromHiddenSet := true ].

"Must set indexedPaths first so tracker is not invoked for instance variable
 modification."
self _getIndexList _performNoDebug: #addTracker: with: tracker.

removeFromHiddenSet
  ifTrue: [ System _remove: self from: 23 ].

super _setModificationTrackingTo: tracker.
System _disableProtectedMode.
%
category: 'Updating Indexes - Private'
method: UnorderedCollection
_undoIndexCreation: indexObj pathTerm: pathTerm

"Removes any vestiges of the incomplete index, and attempts to commit this
 transaction until successful."

| parentTerm iList |
[
  System abortTransaction.

  parentTerm := pathTerm getParentTerm.

  " remove index from each path term "
  1 to: indexObj size do: [ :i | (indexObj at: i) removeIndex: indexObj ].

  " make path term (and its children) unusable "
  pathTerm _makeObsolete.

  " remove the entry in the receiver's index list "
  (iList := self _indexedPaths) ~~ nil
    ifTrue: [ iList removeIndex: indexObj withOffset: 1 for: self ].

  parentTerm ~~ nil
    ifTrue: [ parentTerm _determineChildren ].

  System commitTransaction.
] untilTrue.
%
category: 'Adding'
method: UnorderedCollection
_updateIndexesForAdditionOf: anObject logging: aBoolean

"anObject is being added to the receiver.  Update any indexes if necessary.
 Returns true if the index objects were modified correctly; otherwise returns
 an Array containing error information."

<primitive: 901>
| mapInfo rootTerms pathTerm doLogging iList aSet
 i sz indexObj offset num val prevTerm indexDict hasIndex |

iList := self _indexedPaths.
" first handle modification tracking "
hasIndex := false.
1 to: iList size by: 2 do: [ :j |
  (iList at: j + 1) > 0
    ifTrue: [ hasIndex := true ]
    ifFalse: [
      (iList at: j + 1) == 0
        ifTrue: [ (iList at: j) adding: anObject to: self ]
    ]
].

hasIndex
  ifFalse: [
    System _disableProtectedMode.
    ^ true
  ].

" if the indexes are not committed, do not need to log "
doLogging := iList isCommitted _and: [ UnorderedCollection _isRcIndexLoggingEnabled ].

rootTerms := iList rootTerms.
sz := rootTerms size.

Exception
  category: nil
  number: nil
  do: [ :ex :cat :num :args | | txt |
    " get the text for the raised error "
    txt := cat textForError: num args: args.
    " returns Array containing err info "
    ^ #[ true " indicating indexing objects may have been modified ",
      num,
      txt ]
  ].

"Make a pass through the indexedPaths to see if the receiver participates as a
 set-valued instance variable.  If so, then must update the index dictionary."
1 to: iList size by: 2 do: [ :j |
  indexObj := iList at: j.
  offset := iList at: j + 1.
  (offset > 1 _and: [ indexObj isComplete ])
    ifTrue: [
      " lazy initialization of aSet (in case we don't ever need it) "
      aSet == nil
        ifTrue: [ aSet := IdentitySet new ].

      " get the previous path term (the one with an asterisk) "
      pathTerm := indexObj at: offset - 1.
      " only update the dictionary if we haven't already done it "
      (aSet includesIdentical: pathTerm)
        ifFalse: [
          aSet add: pathTerm.

          indexDict := indexObj indexDictionary.
          " see how many entries exist "
          val := indexDict
            at: self
            term: pathTerm getParentTerm
            otherwise: nil.

          nil == val
            ifTrue: [
              indexDict
                _errorNoEntryForKey: self
                term: pathTerm getParentTerm
                value: nil
            ]
            ifFalse: [
              (BucketValueBag _hasInstance: val)
                ifTrue: [ num := val size ]
                ifFalse: [ num := 1 ]
            ].

          num timesRepeat: [
            " add a dictionary entry for anObject -> NSC "
            indexDict
              _at: anObject
              put: self
              term: pathTerm
              logging: doLogging
          ]
        ]
    ].
].

" now make a pass through the root path terms "
i := 1.
[ i <= sz _and: [ i <= rootTerms size ] ] whileTrue: [
  pathTerm := rootTerms at: i.

  pathTerm _isObsoletePathTerm
    ifFalse: [

      " see if the receiver is participating as a set-valued instance variable "
      pathTerm offset == 1
        ifTrue: [ num := 1 ]
        ifFalse: [
          " get the path term before the set-valued path term "
          prevTerm := pathTerm getParentTerm.
          pathTerm indicatesIndexOnNscElements
            ifFalse: [ prevTerm := prevTerm getParentTerm ].

          val := pathTerm getIndexDictionary
            at: self
            term: prevTerm
            otherwise: nil.
          " if not found, then mappings have already been added "
          nil == val
            ifTrue: [ num := 0 ]
            ifFalse: [
              (BucketValueBag _hasInstance: val)
                ifTrue: [ num := val size ]
                ifFalse: [ num := 1 ]
            ]
        ].

        " if the index is on elements of the NSC, go ahead and update the indexes "
        pathTerm indicatesIndexOnNscElements
          ifTrue: [
            pathTerm updateBtree ~~ nil
              ifTrue: [
                (pathTerm _checkBtreeComparisonWith: anObject)
                  ifFalse: [
                    i > 1
                      ifTrue: [
                        " indexing objects have been modified, prevent commits "
                        anObject _error: #rtErrRangeEqualityIndexInvalidClassKindForBtree
                          args: #[ pathTerm updateBtree lastElementClass ]
                      ]
                      ifFalse: [
                        System _disableProtectedMode.
                        ^ #[ false,  " indicating no indexing objects were modified "
                          (ErrorSymbols at: #rtErrRangeEqualityIndexInvalidClassKindForBtree),
                          anObject,
                          pathTerm updateBtree lastElementClass ]
                      ]
                  ]
              ].

            num timesRepeat: [
              pathTerm addDirectMappingFor: anObject logging: doLogging
            ]
          ]
          ifFalse: [
            ( pathTerm isConstrained _or:
            [ pathTerm _doNotPerformPretraversalFor: anObject ] )
              " if the path is constrained, go ahead and update the indexes "
              ifTrue: [
                num timesRepeat: [
                  pathTerm addMappingsForObject: anObject logging: doLogging
                ]
              ]
              ifFalse: [
                " get all the mapping info first (this detects any errors
                  along the path before any changes are made to the indexes) "
                mapInfo := pathTerm
                  getMappingInfoFor: anObject
                  ifObject: nil
                  atOffset: 0
                  replaceWith: nil.

                " if the result is not a map info object, it is an Array
                  used for error information "
                (mapInfo _class == MappingInfo)
                  ifFalse: [
                    i > 1
                      ifTrue: [ | errArgs |
                        " indexing objects have been modified, prevent commits "
                        errArgs := Array new.
                        3 to: mapInfo size do: [ :j | errArgs add: (mapInfo at: j) ].
                        System signal: (mapInfo at: 2) args: errArgs signalDictionary: GemStoneError.
                      ].
                    System _disableProtectedMode.
                    ^ mapInfo
                  ].

                " now update dependency lists, index dictionary, B-trees "
                num timesRepeat: [
                  mapInfo pathTerm addMappingsUsing: mapInfo logging: doLogging
                ]
            ]
        ]
    ].
  i := i + 1
].

" update the index list's segments set "
iList addSegment: anObject segment logging: doLogging.

System _disableProtectedMode.
^ true
%
category: 'Removing'
method: UnorderedCollection
_updateIndexesForRemovalOf: anObject

"anObject is being removed from the receiver.  Update the indexes if
 necessary.  This method is invoked prior to removal of anObject
 from the receiver. "

<primitive: 901>
| depList pathTerm rootTerms lastOne iList aSet sz i indexObj offset
currSize origSize prevTerm val num doLogging hasIndex |

" if it is not an RcIdentityBag, anObject may not be in the receiver "
( self class isNsc _and: [ (self includes: anObject) not ] )
  ifTrue: [
    System _disableProtectedMode.
    ^ self
  ].

iList := self _indexedPaths.
" first handle modification tracking "
hasIndex := false.
1 to: iList size by: 2 do: [ :j |
  (iList at: j + 1) > 0
    ifTrue: [ hasIndex := true ]
    ifFalse: [
      (iList at: j + 1) == 0
        ifTrue: [ (iList at: j) removing: anObject from: self ]
    ]
].

hasIndex
  ifFalse: [
    System _disableProtectedMode.
    ^ self
  ].

Exception
  category: nil
  number: nil
  do: [ :ex :cat :num :args | | txt |
  " get the text for the raised error "
  txt := cat textForError: num args: args.
  " check for recursive signal "
  num == (ErrorSymbols at: #rtErrPreventingCommit)
    ifTrue: [ " remove this exception and resignal "
      ex remove.
      ex resignal: cat number: num args: args
    ]
    ifFalse: [ " append additional message to the end of text "
      txt _error: #rtErrPreventingCommit
    ]
  ].

" get the removed object's dependency list "
depList := DependencyList for: anObject.

" see if another occurrence of anObject is in the index objects "
lastOne := self _isLastOccurrenceInIndexObjectsFor: anObject.

doLogging := UnorderedCollection _isRcIndexLoggingEnabled.
origSize := iList size.

rootTerms := iList rootTerms.
sz := rootTerms size.
i := 1.
" for each unique path on which there is an index "
[ i <= sz _and: [ i <= rootTerms size ] ] whileTrue: [
  pathTerm := rootTerms at: i.
  " see if the path term is still a root term and is not obsolete "
  ((iList rootTerms includesIdentical: pathTerm) _and:
  [ pathTerm _isObsoletePathTerm not ])
    ifTrue: [
      " see if the receiver is participating as a set-valued instance variable "
      pathTerm offset == 1
        ifTrue: [ num := 1 ]
        ifFalse: [
          " get the path term before the set-valued path term "
          prevTerm := pathTerm getParentTerm.
          pathTerm indicatesIndexOnNscElements
            ifFalse: [ prevTerm := prevTerm getParentTerm ].

          val := pathTerm getIndexDictionary
            at: self
            term: prevTerm
            otherwise: nil.
          " if not found, then mappings have already been removed "
          nil == val
            ifTrue: [ num := 0 ]
            ifFalse: [
              (BucketValueBag _hasInstance: val)
                ifTrue: [ num := val size ]
                ifFalse: [ num := 1 ]
            ]
        ].

      " see if this index is on the elements of the NSC "
      pathTerm indicatesIndexOnNscElements
        ifTrue: [
          num timesRepeat: [
            pathTerm removeDirectMappingFor: anObject logging: doLogging.
          ].

          ( depList ~~ nil _and: [ pathTerm needsDepList ] )
            ifTrue: [ DependencyList removePathTerm: pathTerm for: anObject ]
        ]
        ifFalse: [
          nil == anObject
            ifFalse: [ 
              num - 1 timesRepeat: [
                pathTerm removeMappingsFor: anObject
                  lastOne: false
                  logging: doLogging
              ].
              pathTerm removeMappingsFor: anObject
                lastOne: lastOne
                logging: doLogging
            ].
          ( depList ~~ nil _and: [ lastOne ] )
            ifTrue: [ DependencyList removePathTerm: pathTerm for: anObject ]
        ].
    ].
  i := i + 1
].

currSize := iList size.
i := 1.
[ i <= origSize _and: [ i <= currSize ] ] whileTrue: [
  indexObj := iList at: i.
  offset := iList at: i + 1.
  " see if receiver participates as a set-valued instance variable "
  (offset > 1 _and: [ indexObj isComplete ])
    ifTrue: [
      " lazy initialization of aSet (in case we don't ever need it) "
      aSet == nil
        ifTrue: [ aSet := IdentitySet new ].
      " get the previous path term (the one with an asterisk) "
      pathTerm := indexObj at: offset - 1.
      " only update the dictionary if we haven't already done it "
      (aSet includesIdentical: pathTerm)
        ifFalse: [
          aSet add: pathTerm.
          " see how many entries exist "
          val := indexObj indexDictionary
            at: self
            term: pathTerm getParentTerm
            otherwise: nil.

          " if not found, then mappings have already been removed "
          nil == val
            ifTrue: [ num := 0 ]
            ifFalse: [
              (BucketValueBag _hasInstance: val)
                ifTrue: [ num := val size ]
                ifFalse: [ num := 1 ]
            ].

          num timesRepeat: [
            " remove dictionary entry for anObject -> NSC "
            indexObj indexDictionary
              removeKey: anObject
              value: self
              term: pathTerm
              logging: doLogging
          ]
        ]
    ].
  i := i + 2
].
System _disableProtectedMode.
%
category: 'Auditing'
method: UnorderedCollection
repairInternalStructures
"For an explanation of this method, refer to the comments in the method
 UnorderedCollection>>invalidReferenceAuditWithRepair:"
 
^self invalidReferenceAuditWithRepair: true
%
category: 'Auditing'
method: UnorderedCollection
auditInternalStructuresWithRepair: aBoolean
"Audit and optionally repair the internal structures of the receiver.
 If aBoolean is true, errors will be fixed as they are detected, but
 the transaction is not automatically committed and a commit after this
 method returns is required.  If aBoolean is false, the audit is 
 read-only operation and no objects are modified.  Invalid reference 
 errors found and repaired by the invalidReferenceAuditWithRepair:
 method are also detected and repaired by this method.

 Answers a Boolean indicating if the collection is free of errors. 
 true means the collection passed the audit, false means the audit
 detected one or more errors.

 This primitive prints many messages to stdout and is intended to
 be run from a linked topaz session."

<primitive: 552>
^self _primitiveFailed: #auditInternalStructuresWithRepair:
%
category: 'Auditing'
method: UnorderedCollection
auditInternalStructures
"See the method auditInternalStructuresWithRepair: aBoolean for
 more information on this method."

^self auditInternalStructuresWithRepair: false
%
